# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Distributional modules: these are modules that return a tfd.Distribution.

There are useful modules in `acme.networks.stochastic` to either sample or
take the mean of these distributions.
"""

import types
from typing import Optional, Union
from absl import logging
from acme.tf.networks import distributions as ad
import numpy as np
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions
snt_init = snt.initializers

_MIN_SCALE = 1e-4


class DiscreteValuedHead(snt.Module):
  """Represents a parameterized discrete valued distribution.

  The returned distribution is essentially a `tfd.Categorical`, but one which
  knows its support and so can compute the mean value.
  """

  def __init__(self,
               vmin: Union[float, np.ndarray, tf.Tensor],
               vmax: Union[float, np.ndarray, tf.Tensor],
               num_atoms: int,
               w_init: Optional[snt.initializers.Initializer] = None,
               b_init: Optional[snt.initializers.Initializer] = None):
    """Initialization.

    If vmin and vmax have shape S, this will store the category values as a
    Tensor of shape (S*, num_atoms).

    Args:
      vmin: Minimum of the value range
      vmax: Maximum of the value range
      num_atoms: The atom values associated with each bin.
      w_init: Initialization for linear layer weights.
      b_init: Initialization for linear layer biases.
    """
    super().__init__(name='DiscreteValuedHead')
    vmin = tf.convert_to_tensor(vmin)
    vmax = tf.convert_to_tensor(vmax)
    self._values = tf.linspace(vmin, vmax, num_atoms, axis=-1)
    self._distributional_layer = snt.Linear(tf.size(self._values),
                                            w_init=w_init,
                                            b_init=b_init)

  def __call__(self, inputs: tf.Tensor) -> tfd.Distribution:
    logits = self._distributional_layer(inputs)
    logits = tf.reshape(logits,
                        tf.concat([tf.shape(logits)[:1],  # batch size
                                   tf.shape(self._values)],
                                  axis=0))
    values = tf.cast(self._values, logits.dtype)

    return ad.DiscreteValuedDistribution(values=values, logits=logits)


class MultivariateNormalDiagHead(snt.Module):
  """Module that produces a multivariate normal distribution using tfd.Independent or tfd.MultivariateNormalDiag."""

  def __init__(
      self,
      num_dimensions: int,
      init_scale: float = 0.3,
      min_scale: float = 1e-6,
      tanh_mean: bool = False,
      fixed_scale: bool = False,
      use_tfd_independent: bool = False,
      w_init: snt_init.Initializer = tf.initializers.VarianceScaling(1e-4),
      b_init: snt_init.Initializer = tf.initializers.Zeros()):
    """Initialization.

    Args:
      num_dimensions: Number of dimensions of MVN distribution.
      init_scale: Initial standard deviation.
      min_scale: Minimum standard deviation.
      tanh_mean: Whether to transform the mean (via tanh) before passing it to
        the distribution.
      fixed_scale: Whether to use a fixed variance.
      use_tfd_independent: Whether to use tfd.Independent or
        tfd.MultivariateNormalDiag class
      w_init: Initialization for linear layer weights.
      b_init: Initialization for linear layer biases.
    """
    super().__init__(name='MultivariateNormalDiagHead')
    self._init_scale = init_scale
    self._min_scale = min_scale
    self._tanh_mean = tanh_mean
    self._mean_layer = snt.Linear(num_dimensions, w_init=w_init, b_init=b_init)
    self._fixed_scale = fixed_scale

    if not fixed_scale:
      self._scale_layer = snt.Linear(
          num_dimensions, w_init=w_init, b_init=b_init)
    self._use_tfd_independent = use_tfd_independent

  def __call__(self, inputs: tf.Tensor) -> tfd.Distribution:
    zero = tf.constant(0, dtype=inputs.dtype)
    mean = self._mean_layer(inputs)

    if self._fixed_scale:
      scale = tf.ones_like(mean) * self._init_scale
    else:
      scale = tf.nn.softplus(self._scale_layer(inputs))
      scale *= self._init_scale / tf.nn.softplus(zero)
      scale += self._min_scale

    # Maybe transform the mean.
    if self._tanh_mean:
      mean = tf.tanh(mean)

    if self._use_tfd_independent:
      dist = tfd.Independent(tfd.Normal(loc=mean, scale=scale))
    else:
      dist = tfd.MultivariateNormalDiag(loc=mean, scale_diag=scale)

    return dist


class GaussianMixture(snt.Module):
  """Module that outputs a Gaussian Mixture Distribution."""

  def __init__(self,
               num_dimensions: int,
               num_components: int,
               multivariate: bool,
               init_scale: Optional[float] = None,
               name: str = 'GaussianMixture'):
    """Initialization.

    Args:
      num_dimensions: dimensionality of the output distribution
      num_components: number of mixture components.
      multivariate: whether the resulting distribution is multivariate or not.
      init_scale: the initial scale for the Gaussian mixture components.
      name: name of the module passed to snt.Module parent class.
    """
    super().__init__(name=name)

    self._num_dimensions = num_dimensions
    self._num_components = num_components
    self._multivariate = multivariate

    if init_scale is not None:
      self._scale_factor = init_scale / tf.nn.softplus(0.)
    else:
      self._scale_factor = 1.0  # Corresponds to init_scale = softplus(0).

    # Define the weight initializer.
    w_init = tf.initializers.VarianceScaling(1e-5)

    # Create a layer that outputs the unnormalized log-weights.
    if self._multivariate:
      logits_size = self._num_components
    else:
      logits_size = self._num_dimensions * self._num_components
    self._logit_layer = snt.Linear(logits_size, w_init=w_init)

    # Create two layers that outputs a location and a scale, respectively, for
    # each dimension and each component.
    self._loc_layer = snt.Linear(
        self._num_dimensions * self._num_components, w_init=w_init)
    self._scale_layer = snt.Linear(
        self._num_dimensions * self._num_components, w_init=w_init)

  def __call__(self,
               inputs: tf.Tensor,
               low_noise_policy: bool = False) -> tfd.Distribution:
    """Run the networks through inputs.

    Args:
      inputs: hidden activations of the policy network body.
      low_noise_policy: whether to set vanishingly small scales for each
        component. If this flag is set to True, the policy is effectively run
        without Gaussian noise.

    Returns:
      Mixture Gaussian distribution.
    """

    # Compute logits, locs, and scales if necessary.
    logits = self._logit_layer(inputs)
    locs = self._loc_layer(inputs)

    # When a low_noise_policy is requested, set the scales to its minimum value.
    if low_noise_policy:
      scales = tf.fill(locs.shape, _MIN_SCALE)
    else:
      scales = self._scale_layer(inputs)
      scales = self._scale_factor * tf.nn.softplus(scales) + _MIN_SCALE

    if self._multivariate:
      shape = [-1, self._num_components, self._num_dimensions]
      # Reshape the mixture's location and scale parameters appropriately.
      locs = tf.reshape(locs, shape)
      scales = tf.reshape(scales, shape)
      # In this case, no need to reshape logits as they are in the correct shape
      # already, namely [batch_size, num_components].
      components_distribution = tfd.MultivariateNormalDiag(
          loc=locs, scale_diag=scales)
    else:
      shape = [-1, self._num_dimensions, self._num_components]
      # Reshape the mixture's location and scale parameters appropriately.
      locs = tf.reshape(locs, shape)
      scales = tf.reshape(scales, shape)
      components_distribution = tfd.Normal(loc=locs, scale=scales)
      logits = tf.reshape(logits, shape)

    # Create the mixture distribution.
    distribution = tfd.MixtureSameFamily(
        mixture_distribution=tfd.Categorical(logits=logits),
        components_distribution=components_distribution)

    if not self._multivariate:
      distribution = tfd.Independent(distribution)

    return distribution


class UnivariateGaussianMixture(GaussianMixture):
  """Head which outputs a Mixture of Gaussians Distribution."""

  def __init__(self,
               num_dimensions: int,
               num_components: int = 5,
               init_scale: Optional[float] = None,
               num_mixtures: Optional[int] = None):
    """Create an mixture of Gaussian actor head.

    Args:
      num_dimensions: dimensionality of the output distribution. Each dimension
        is going to be an independent 1d GMM model.
      num_components: number of mixture components.
      init_scale: the initial scale for the Gaussian mixture components.
      num_mixtures: deprecated argument which overwrites num_components.
    """
    if num_mixtures is not None:
      logging.warning("""the num_mixtures parameter has been deprecated; use
                    num_components instead; the value of num_components is being
                    ignored""")
      num_components = num_mixtures
    super().__init__(num_dimensions=num_dimensions,
                     num_components=num_components,
                     multivariate=False,
                     init_scale=init_scale,
                     name='UnivariateGaussianMixture')


class MultivariateGaussianMixture(GaussianMixture):
  """Head which outputs a mixture of multivariate Gaussians distribution."""

  def __init__(self,
               num_dimensions: int,
               num_components: int = 5,
               init_scale: Optional[float] = None):
    """Initialization.

    Args:
      num_dimensions: dimensionality of the output distribution
        (also the dimensionality of the multivariate Gaussian model).
      num_components: number of mixture components.
      init_scale: the initial scale for the Gaussian mixture components.
    """
    super().__init__(num_dimensions=num_dimensions,
                     num_components=num_components,
                     multivariate=True,
                     init_scale=init_scale,
                     name='MultivariateGaussianMixture')


class ApproximateMode(snt.Module):
  """Override the mode function of the distribution.

  For non-constant Jacobian transformed distributions, the mode is non-trivial
  to compute, so for these distributions the mode function is not supported in
  TFP. A frequently used approximation is to forward transform the mode of the
  untransformed distribution.

  Otherwise (an untransformed distribution or a transformed distribution with a
  constant Jacobian), this is a no-op.
  """

  def __call__(self, inputs: tfd.Distribution) -> tfd.Distribution:
    if isinstance(inputs, tfd.TransformedDistribution):
      if not inputs.bijector.is_constant_jacobian:
        def _mode(self, **kwargs):
          distribution_kwargs, bijector_kwargs = self._kwargs_split_fn(kwargs)
          x = self.distribution.mode(**distribution_kwargs)
          y = self.bijector.forward(x, **bijector_kwargs)
          return y
        inputs._mode = types.MethodType(_mode, inputs)
    return inputs



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme.utils.loggers.base."""

from absl.testing import absltest
from acme.utils.loggers import base
import jax.numpy as jnp
import numpy as np
import tensorflow as tf


class BaseTest(absltest.TestCase):

  def test_tensor_serialisation(self):
    data = {'x': tf.zeros(shape=(32,))}
    output = base.to_numpy(data)
    expected = {'x': np.zeros(shape=(32,))}
    np.testing.assert_array_equal(output['x'], expected['x'])

  def test_device_array_serialisation(self):
    data = {'x': jnp.zeros(shape=(32,))}
    output = base.to_numpy(data)
    expected = {'x': np.zeros(shape=(32,))}
    np.testing.assert_array_equal(output['x'], expected['x'])


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""IMPALA Builder."""
from typing import Any, Callable, Iterator, List, Optional

import acme
from acme import adders
from acme import core
from acme import specs
from acme.adders import reverb as reverb_adders
from acme.agents.jax import builders
from acme.agents.jax.impala import acting
from acme.agents.jax.impala import config as impala_config
from acme.agents.jax.impala import learning
from acme.agents.jax.impala import networks as impala_networks
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import haiku as hk
import jax.numpy as jnp
import optax
import reverb


class IMPALABuilder(builders.ActorLearnerBuilder):
  """IMPALA Builder."""

  def __init__(
      self,
      config: impala_config.IMPALAConfig,
      core_state_spec: hk.LSTMState,
      logger_fn: Callable[[], loggers.Logger] = lambda: None,
      table_extension: Optional[Callable[[], Any]] = None,
  ):
    """Creates an IMPALA learner."""
    self._config = config
    self._core_state_spec = core_state_spec
    self._sequence_length = self._config.sequence_length
    self._num_sequences_per_batch = self._config.batch_size
    self._logger_fn = logger_fn
    self._table_extension = table_extension

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    """The queue; use XData or INFO log."""

    num_actions = environment_spec.actions.num_values
    extra_spec = {
        'core_state': self._core_state_spec,
        'logits': jnp.ones(shape=(num_actions,), dtype=jnp.float32)
    }
    signature = reverb_adders.SequenceAdder.signature(
        environment_spec,
        extra_spec,
        sequence_length=self._config.sequence_length)

    # Maybe create rate limiter.
    # Setting the samples_per_insert ratio less than the default of 1.0, allows
    # the agent to drop data for the benefit of using data from most up-to-date
    # policies to compute its learner updates.
    samples_per_insert = self._config.samples_per_insert
    if samples_per_insert:
      if samples_per_insert > 1.0 or samples_per_insert <= 0.0:
        raise ValueError(
            'Impala requires a samples_per_insert ratio in the range (0, 1],'
            f' but received {samples_per_insert}.')
      limiter = reverb.rate_limiters.SampleToInsertRatio(
          samples_per_insert=samples_per_insert,
          min_size_to_sample=1,
          error_buffer=self._config.batch_size)
    else:
      limiter = reverb.rate_limiters.MinSize(1)

    table_extensions = []
    if self._table_extension is not None:
      table_extensions = [self._table_extension()]
    queue = reverb.Table(
        name=self._config.replay_table_name,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=self._config.max_queue_size,
        max_times_sampled=1,
        rate_limiter=limiter,
        extensions=table_extensions,
        signature=signature)
    return [queue]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Creates a dataset."""
    # NOTE: we need TrajectoryDataset for SequenceAdder.
    # NOTE: Value for max_in_flight_samples_per_worker comes from a
    # reverb.ReplayDataset recommendation of 2-3x batch size.
    dataset = reverb.TrajectoryDataset.from_table_signature(
        server_address=replay_client.server_address,
        table=self._config.replay_table_name,
        max_in_flight_samples_per_worker=2 * self._config.batch_size)

    # Add batch dimension.
    dataset = dataset.batch(self._num_sequences_per_batch, drop_remainder=True)
    return dataset.as_numpy_iterator()

  def make_adder(self, replay_client: reverb.Client) -> adders.Adder:
    """Creates an adder which handles observations."""
    # Note that the last transition in the sequence is used for bootstrapping
    # only and is ignored otherwise. So we need to make sure that sequences
    # overlap on one transition, thus "-1" in the period length computation.
    return reverb_adders.SequenceAdder(
        client=replay_client,
        priority_fns={self._config.replay_table_name: None},
        period=self._config.sequence_period or (self._sequence_length - 1),
        sequence_length=self._sequence_length,
        pad_end_of_episode=False,
        break_end_of_episode=False,
    )

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: impala_networks.IMPALANetworks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:

    optimizer = optax.chain(
        optax.clip_by_global_norm(self._config.max_gradient_norm),
        optax.adam(
            self._config.learning_rate,
            b1=self._config.adam_momentum_decay,
            b2=self._config.adam_variance_decay),
    )

    return learning.IMPALALearner(
        networks=networks,
        iterator=dataset,
        optimizer=optimizer,
        random_key=random_key,
        discount=self._config.discount,
        entropy_cost=self._config.entropy_cost,
        baseline_cost=self._config.baseline_cost,
        max_abs_reward=self._config.max_abs_reward,
        counter=counter,
        logger=self._logger_fn(),
    )

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None) -> acme.Actor:
    variable_client = variable_utils.VariableClient(
        client=variable_source, key='network', device='cpu')
    return acting.IMPALAActor(
        forward_fn=policy_network.forward_fn,
        initial_state_init_fn=policy_network.initial_state_init_fn,
        initial_state_fn=policy_network.initial_state_fn,
        variable_client=variable_client,
        adder=adder,
        rng=hk.PRNGSequence(random_key),
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Networks used in discrete-action agents."""

import sonnet as snt
import tensorflow as tf


class DiscreteFilteredQNetwork(snt.Module):
  """Discrete filtered Q-network.

  This produces filtered Q values according to the method used in the discrete
  BCQ algorithm (https://arxiv.org/pdf/1910.01708.pdf - section 4).
  """

  def __init__(self,
               g_network: snt.Module,
               q_network: snt.Module,
               threshold: float):
    super().__init__(name='discrete_filtered_qnet')
    assert threshold >= 0 and threshold <= 1
    self.g_network = g_network
    self.q_network = q_network
    self._threshold = threshold

  def __call__(self, o_t: tf.Tensor) -> tf.Tensor:
    q_t = self.q_network(o_t)
    g_t = tf.nn.softmax(self.g_network(o_t))
    normalized_g_t = g_t / tf.reduce_max(g_t, axis=-1, keepdims=True)

    # Filter actions based on g_network outputs.
    min_q = tf.reduce_min(q_t, axis=-1, keepdims=True)
    return tf.where(normalized_g_t >= self._threshold, q_t, min_q)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ValueDice learner implementation."""

import functools
import time
from typing import Any, Dict, Iterator, List, Mapping, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.agents.jax.value_dice import networks as value_dice_networks
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax
import reverb


class TrainingState(NamedTuple):
  """Contains training state for the learner."""
  policy_optimizer_state: optax.OptState
  policy_params: networks_lib.Params
  nu_optimizer_state: optax.OptState
  nu_params: networks_lib.Params
  key: jnp.ndarray
  steps: int


def _orthogonal_regularization_loss(params: networks_lib.Params):
  """Orthogonal regularization.

  See equation (3) in https://arxiv.org/abs/1809.11096.

  Args:
    params: Dictionary of parameters to apply regualization for.

  Returns:
    A regularization loss term.
  """
  reg_loss = 0
  for key in params:
    if isinstance(params[key], Mapping):
      reg_loss += _orthogonal_regularization_loss(params[key])
      continue
    variable = params[key]
    assert len(variable.shape) in [1, 2, 4]
    if len(variable.shape) == 1:
      # This is a bias so do not apply regularization.
      continue
    if len(variable.shape) == 4:
      # CNN
      variable = jnp.reshape(variable, (-1, variable.shape[-1]))
    prod = jnp.matmul(jnp.transpose(variable), variable)
    reg_loss += jnp.sum(jnp.square(prod * (1 - jnp.eye(prod.shape[0]))))
  return reg_loss


class ValueDiceLearner(acme.Learner):
  """ValueDice learner."""

  _state: TrainingState

  def __init__(self,
               networks: value_dice_networks.ValueDiceNetworks,
               policy_optimizer: optax.GradientTransformation,
               nu_optimizer: optax.GradientTransformation,
               discount: float,
               rng: jnp.ndarray,
               iterator_replay: Iterator[reverb.ReplaySample],
               iterator_demonstrations: Iterator[types.Transition],
               alpha: float = 0.05,
               policy_reg_scale: float = 1e-4,
               nu_reg_scale: float = 10.0,
               num_sgd_steps_per_step: int = 1,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None):

    rng, policy_key, nu_key = jax.random.split(rng, 3)
    policy_init_params = networks.policy_network.init(policy_key)
    policy_optimizer_state = policy_optimizer.init(policy_init_params)

    nu_init_params = networks.nu_network.init(nu_key)
    nu_optimizer_state = nu_optimizer.init(nu_init_params)

    def compute_losses(
        policy_params: networks_lib.Params,
        nu_params: networks_lib.Params,
        key: jnp.ndarray,
        replay_o_tm1: types.NestedArray,
        replay_a_tm1: types.NestedArray,
        replay_o_t: types.NestedArray,
        demo_o_tm1: types.NestedArray,
        demo_a_tm1: types.NestedArray,
        demo_o_t: types.NestedArray,
    ) -> jnp.ndarray:
      # TODO(damienv, hussenot): what to do with the discounts ?

      def policy(obs, key):
        dist_params = networks.policy_network.apply(policy_params, obs)
        return networks.sample(dist_params, key)

      key1, key2, key3, key4 = jax.random.split(key, 4)

      # Predicted actions.
      demo_o_t0 = demo_o_tm1
      policy_demo_a_t0 = policy(demo_o_t0, key1)
      policy_demo_a_t = policy(demo_o_t, key2)
      policy_replay_a_t = policy(replay_o_t, key3)

      replay_a_tm1 = networks.encode_action(replay_a_tm1)
      demo_a_tm1 = networks.encode_action(demo_a_tm1)
      policy_demo_a_t0 = networks.encode_action(policy_demo_a_t0)
      policy_demo_a_t = networks.encode_action(policy_demo_a_t)
      policy_replay_a_t = networks.encode_action(policy_replay_a_t)

      # "Value function" nu over the expert states.
      nu_demo_t0 = networks.nu_network.apply(nu_params, demo_o_t0,
                                             policy_demo_a_t0)
      nu_demo_tm1 = networks.nu_network.apply(nu_params, demo_o_tm1, demo_a_tm1)
      nu_demo_t = networks.nu_network.apply(nu_params, demo_o_t,
                                            policy_demo_a_t)
      nu_demo_diff = nu_demo_tm1 - discount * nu_demo_t

      # "Value function" nu over the replay buffer states.
      nu_replay_tm1 = networks.nu_network.apply(nu_params, replay_o_tm1,
                                                replay_a_tm1)
      nu_replay_t = networks.nu_network.apply(nu_params, replay_o_t,
                                              policy_replay_a_t)
      nu_replay_diff = nu_replay_tm1 - discount * nu_replay_t

      # Linear part of the loss.
      linear_loss_demo = jnp.mean(nu_demo_t0 * (1.0 - discount))
      linear_loss_rb = jnp.mean(nu_replay_diff)
      linear_loss = (linear_loss_demo * (1 - alpha) + linear_loss_rb * alpha)

      # Non linear part of the loss.
      nu_replay_demo_diff = jnp.concatenate([nu_demo_diff, nu_replay_diff],
                                            axis=0)
      replay_demo_weights = jnp.concatenate([
          jnp.ones_like(nu_demo_diff) * (1 - alpha),
          jnp.ones_like(nu_replay_diff) * alpha
      ],
                                            axis=0)
      replay_demo_weights /= jnp.mean(replay_demo_weights)
      non_linear_loss = jnp.sum(
          jax.lax.stop_gradient(
              utils.weighted_softmax(nu_replay_demo_diff, replay_demo_weights,
                                     axis=0)) *
          nu_replay_demo_diff)

      # Final loss.
      loss = (non_linear_loss - linear_loss)

      # Regularized policy loss.
      if policy_reg_scale > 0.:
        policy_reg = _orthogonal_regularization_loss(policy_params)
      else:
        policy_reg = 0.

      # Gradient penality on nu
      if nu_reg_scale > 0.0:
        batch_size = demo_o_tm1.shape[0]
        c = jax.random.uniform(key4, shape=(batch_size,))
        shape_o = [
            dim if i == 0 else 1 for i, dim in enumerate(replay_o_tm1.shape)
        ]
        shape_a = [
            dim if i == 0 else 1 for i, dim in enumerate(replay_a_tm1.shape)
        ]
        c_o = jnp.reshape(c, shape_o)
        c_a = jnp.reshape(c, shape_a)
        mixed_o_tm1 = c_o * demo_o_tm1 + (1 - c_o) * replay_o_tm1
        mixed_a_tm1 = c_a * demo_a_tm1 + (1 - c_a) * replay_a_tm1
        mixed_o_t = c_o * demo_o_t + (1 - c_o) * replay_o_t
        mixed_policy_a_t = c_a * policy_demo_a_t + (1 - c_a) * policy_replay_a_t
        mixed_o = jnp.concatenate([mixed_o_tm1, mixed_o_t], axis=0)
        mixed_a = jnp.concatenate([mixed_a_tm1, mixed_policy_a_t], axis=0)

        def sum_nu(o, a):
          return jnp.sum(networks.nu_network.apply(nu_params, o, a))

        nu_grad_o_fn = jax.grad(sum_nu, argnums=0)
        nu_grad_a_fn = jax.grad(sum_nu, argnums=1)
        nu_grad_o = nu_grad_o_fn(mixed_o, mixed_a)
        nu_grad_a = nu_grad_a_fn(mixed_o, mixed_a)
        nu_grad = jnp.concatenate([
            jnp.reshape(nu_grad_o, [batch_size, -1]),
            jnp.reshape(nu_grad_a, [batch_size, -1])], axis=-1)
        # TODO(damienv, hussenot): check for the need of eps
        # (like in the original value dice code).
        nu_grad_penalty = jnp.mean(
            jnp.square(
                jnp.linalg.norm(nu_grad + 1e-8, axis=-1, keepdims=True) - 1))
      else:
        nu_grad_penalty = 0.0

      policy_loss = -loss + policy_reg_scale * policy_reg
      nu_loss = loss + nu_reg_scale * nu_grad_penalty

      return policy_loss, nu_loss

    def sgd_step(
        state: TrainingState,
        data: Tuple[types.Transition, types.Transition]
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:
      replay_transitions, demo_transitions = data
      key, key_loss = jax.random.split(state.key)
      compute_losses_with_input = functools.partial(
          compute_losses,
          replay_o_tm1=replay_transitions.observation,
          replay_a_tm1=replay_transitions.action,
          replay_o_t=replay_transitions.next_observation,
          demo_o_tm1=demo_transitions.observation,
          demo_a_tm1=demo_transitions.action,
          demo_o_t=demo_transitions.next_observation,
          key=key_loss)
      (policy_loss_value, nu_loss_value), vjpfun = jax.vjp(
          compute_losses_with_input,
          state.policy_params, state.nu_params)
      policy_gradients, _ = vjpfun((1.0, 0.0))
      _, nu_gradients = vjpfun((0.0, 1.0))

      # Update optimizers.
      policy_update, policy_optimizer_state = policy_optimizer.update(
          policy_gradients, state.policy_optimizer_state)
      policy_params = optax.apply_updates(state.policy_params, policy_update)

      nu_update, nu_optimizer_state = nu_optimizer.update(
          nu_gradients, state.nu_optimizer_state)
      nu_params = optax.apply_updates(state.nu_params, nu_update)

      new_state = TrainingState(
          policy_optimizer_state=policy_optimizer_state,
          policy_params=policy_params,
          nu_optimizer_state=nu_optimizer_state,
          nu_params=nu_params,
          key=key,
          steps=state.steps + 1,
      )

      metrics = {
          'policy_loss': policy_loss_value,
          'nu_loss': nu_loss_value,
      }

      return new_state, metrics

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(
        'learner', asynchronous=True, serialize_fn=utils.fetch_devicearray)

    # Iterator on demonstration transitions.
    self._iterator_demonstrations = iterator_demonstrations
    self._iterator_replay = iterator_replay

    self._sgd_step = jax.jit(utils.process_multiple_batches(
        sgd_step, num_sgd_steps_per_step))

    # Create initial state.
    self._state = TrainingState(
        policy_optimizer_state=policy_optimizer_state,
        policy_params=policy_init_params,
        nu_optimizer_state=nu_optimizer_state,
        nu_params=nu_init_params,
        key=rng,
        steps=0,
    )

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  def step(self):
    # Get data from replay (dropping extras if any). Note there is no
    # extra data here because we do not insert any into Reverb.
    # TODO(raveman): Add a support for offline training, where we do not consume
    # data from the replay buffer.
    sample = next(self._iterator_replay)
    replay_transitions = types.Transition(*sample.data)

    # Get a batch of Transitions from the demonstration.
    demonstration_transitions = next(self._iterator_demonstrations)

    self._state, metrics = self._sgd_step(
        self._state, (replay_transitions, demonstration_transitions))

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Increment counts and record the current time
    counts = self._counter.increment(steps=1, walltime=elapsed_time)

    # Attempts to write the logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[Any]:
    variables = {
        'policy': self._state.policy_params,
        'nu': self._state.nu_params,
    }
    return [variables[name] for name in names]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Variable handling utilities for TensorFlow 2."""

from concurrent import futures
from typing import Mapping, Optional, Sequence

from acme import core

import tensorflow as tf
import tree


class VariableClient:
  """A variable client for updating variables from a remote source."""

  def __init__(self,
               client: core.VariableSource,
               variables: Mapping[str, Sequence[tf.Variable]],
               update_period: int = 1):
    self._keys = list(variables.keys())
    self._variables = tree.flatten(list(variables.values()))
    self._call_counter = 0
    self._update_period = update_period
    self._client = client
    self._request = lambda: client.get_variables(self._keys)

    # Create a single background thread to fetch variables without necessarily
    # blocking the actor.
    self._executor = futures.ThreadPoolExecutor(max_workers=1)
    self._async_request = lambda: self._executor.submit(self._request)

    # Initialize this client's future to None to indicate to the `update()`
    # method that there is no pending/running request.
    self._future: Optional[futures.Future] = None

  def update(self, wait: bool = False):
    """Periodically updates the variables with the latest copy from the source.

    This stateful update method keeps track of the number of calls to it and,
    every `update_period` call, sends a request to its server to retrieve the
    latest variables.

    If wait is True, a blocking request is executed. Any active request will be
    cancelled.
    If wait is False, this method makes an asynchronous request for variables
    and returns. Unless the request is immediately fulfilled, the variables are
    only copied _within a subsequent call to_ `update()`, whenever the request
    is fulfilled by the `VariableSource`. If there is an existing fulfilled
    request when this method is called, the resulting variables are immediately
    copied.

    Args:
      wait: if True, executes blocking update.
    """
    # Track the number of calls (we only update periodically).
    if self._call_counter < self._update_period:
      self._call_counter += 1

    period_reached: bool = self._call_counter >= self._update_period

    if period_reached and wait:
      # Cancel any active request.
      self._future: Optional[futures.Future] = None
      self.update_and_wait()
      self._call_counter = 0
      return

    if period_reached and self._future is None:
      # The update period has been reached and no request has been sent yet, so
      # making an asynchronous request now.
      self._future = self._async_request()
      self._call_counter = 0

    if self._future is not None and self._future.done():
      # The active request is done so copy the result and remove the future.
      self._copy(self._future.result())
      self._future: Optional[futures.Future] = None
    else:
      # There is either a pending/running request or we're between update
      # periods, so just carry on.
      return

  def update_and_wait(self):
    """Immediately update and block until we get the result."""
    self._copy(self._request())

  def _copy(self, new_variables: Sequence[Sequence[tf.Variable]]):
    """Copies the new variables to the old ones."""

    new_variables = tree.flatten(new_variables)
    if len(self._variables) != len(new_variables):
      raise ValueError('Length mismatch between old variables and new.')

    for new, old in zip(new_variables, self._variables):
      old.assign(new)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local SAC agents, using JAX."""

import functools
from typing import Callable, Optional

from acme import specs
from acme.agents.jax import normalization
from acme.agents.jax.sac import builder
from acme.agents.jax.sac import config as sac_config
from acme.agents.jax.sac import networks
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env


NetworkFactory = Callable[[specs.EnvironmentSpec], networks.SACNetworks]


class DistributedSAC(distributed_layout.DistributedLayout):
  """Distributed program definition for SAC.
  """

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      network_factory: NetworkFactory,
      config: sac_config.SACConfig,
      seed: int,
      num_actors: int,
      max_number_of_steps: Optional[int] = None,
      log_to_bigtable: bool = False,
      log_every: float = 10.0,
      normalize_input: bool = True,
  ):
    logger_fn = functools.partial(loggers.make_default_logger,
                                  'learner', log_to_bigtable,
                                  time_delta=log_every, asynchronous=True,
                                  serialize_fn=utils.fetch_devicearray,
                                  steps_key='learner_steps')
    sac_builder = builder.SACBuilder(config, logger_fn=logger_fn)
    if normalize_input:
      environment_spec = specs.make_environment_spec(environment_factory(False))
      # One batch dimension: [batch_size, ...]
      batch_dims = (0,)
      sac_builder = normalization.NormalizationBuilder(
          sac_builder,
          environment_spec,
          is_sequence_based=False,
          batch_dims=batch_dims)
    eval_policy_factory = (
        lambda n: networks.apply_policy_and_sample(n, True))
    super().__init__(
        seed=seed,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=sac_builder,
        policy_network=networks.apply_policy_and_sample,
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=sac_builder,
                policy_factory=eval_policy_factory,
                log_to_bigtable=log_to_bigtable)
        ],
        num_actors=num_actors,
        max_number_of_steps=max_number_of_steps,
        prefetch_size=config.prefetch_size,
        log_to_bigtable=log_to_bigtable,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            log_to_bigtable, log_every),
    )


class SAC(local_layout.LocalLayout):
  """Local agent for SAC.
  """

  def __init__(
      self,
      spec: specs.EnvironmentSpec,
      network: networks.SACNetworks,
      config: sac_config.SACConfig,
      seed: int,
      normalize_input: bool = True,
      counter: Optional[counting.Counter] = None,
  ):
    min_replay_size = config.min_replay_size
    # Local layout (actually agent.Agent) makes sure that we populate the
    # buffer with min_replay_size initial transitions and that there's no need
    # for tolerance_rate. In order for deadlocks not to happen we need to
    # disable rate limiting that heppens inside the SACBuilder. This is achieved
    # by the following two lines.
    config.samples_per_insert_tolerance_rate = float('inf')
    config.min_replay_size = 1
    sac_builder = builder.SACBuilder(config)
    if normalize_input:
      # One batch dimension: [batch_size, ...]
      batch_dims = (0,)
      sac_builder = normalization.NormalizationBuilder(
          sac_builder, spec, is_sequence_based=False, batch_dims=batch_dims)
    self.builder = sac_builder
    super().__init__(
        seed=seed,
        environment_spec=spec,
        builder=sac_builder,
        networks=network,
        policy_network=networks.apply_policy_and_sample(network),
        batch_size=config.batch_size,
        prefetch_size=config.prefetch_size,
        samples_per_insert=config.samples_per_insert,
        min_replay_size=min_replay_size,
        num_sgd_steps_per_step=config.num_sgd_steps_per_step,
        counter=counter,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local GAIL agents, using JAX.

https://arxiv.org/pdf/1606.03476.pdf
"""

import dataclasses
import functools
from typing import Callable

from acme import specs
from acme.agents.jax import ppo
from acme.agents.jax.ail import agents
from acme.agents.jax.ail import config as ail_config
from acme.agents.jax.ail import losses
from acme.agents.jax.ail import networks as ail_networks
from acme.jax import utils
from acme.utils import loggers
import dm_env


@dataclasses.dataclass
class GAILConfig:
  """Configuration options specific to GAIL."""
  ail_config: ail_config.AILConfig
  ppo_config: ppo.PPOConfig


class DistributedGAIL(agents.DistributedAIL):
  """Distributed program definition for GAIL."""

  def __init__(self,
               environment_factory: Callable[[bool], dm_env.Environment],
               config: GAILConfig,
               *args, **kwargs):
    logger_fn = functools.partial(
        loggers.make_default_logger,
        'direct_learner',
        kwargs['log_to_bigtable'],
        time_delta=kwargs['log_every'],
        asynchronous=True,
        serialize_fn=utils.fetch_devicearray,
        steps_key='learner_steps')
    ppo_agent = ppo.PPOBuilder(
        config.ppo_config, logger_fn=logger_fn)
    kwargs['discriminator_loss'] = losses.gail_loss()
    super().__init__(environment_factory, ppo_agent, config.ail_config, *args,
                     **kwargs)


class GAIL(agents.AIL):
  """Local agent for GAIL."""

  def __init__(self, spec: specs.EnvironmentSpec,
               network: ail_networks.AILNetworks, config: GAILConfig, *args,
               **kwargs):
    ppo_agent = ppo.PPOBuilder(config.ppo_config)
    kwargs['discriminator_loss'] = losses.gail_loss()
    super().__init__(spec, ppo_agent, network, config.ail_config, *args,
                     **kwargs)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of an R2D2 agent."""

from acme.agents.jax.r2d2.agents import DistributedR2D2
from acme.agents.jax.r2d2.agents import DistributedR2D2FromConfig
from acme.agents.jax.r2d2.agents import R2D2
from acme.agents.jax.r2d2.builder import R2D2Builder
from acme.agents.jax.r2d2.config import R2D2Config
from acme.agents.jax.r2d2.learning import R2D2Learner
from acme.agents.jax.r2d2.networks import make_atari_networks
from acme.agents.jax.r2d2.networks import make_behavior_policy
from acme.agents.jax.r2d2.networks import make_networks
from acme.agents.jax.r2d2.networks import R2D2Networks



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test cases used by multiple test files."""

from acme import types
from acme.adders.reverb import sequence as sequence_adder
import dm_env
import numpy as np

# Define the main set of test cases; these are given as parameterized tests to
# the test_adder method and describe a trajectory to add to replay and the
# expected transitions that should result from this trajectory. The expected
# transitions are of the form: (observation, action, reward, discount,
# next_observation, extras).
TEST_CASES_FOR_TRANSITION_ADDER = [
    dict(
        testcase_name='OneStepFinalReward',
        n_step=1,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=0.0, observation=2)),
            (0, dm_env.transition(reward=0.0, observation=3)),
            (0, dm_env.termination(reward=1.0, observation=4)),
        ),
        expected_transitions=(
            types.Transition(1, 0, 0.0, 1.0, 2),
            types.Transition(2, 0, 0.0, 1.0, 3),
            types.Transition(3, 0, 1.0, 0.0, 4),
        )),
    dict(
        testcase_name='OneStepDict',
        n_step=1,
        additional_discount=1.0,
        first=dm_env.restart({'foo': 1}),
        steps=(
            (0, dm_env.transition(reward=0.0, observation={'foo': 2})),
            (0, dm_env.transition(reward=0.0, observation={'foo': 3})),
            (0, dm_env.termination(reward=1.0, observation={'foo': 4})),
        ),
        expected_transitions=(
            types.Transition({'foo': 1}, 0, 0.0, 1.0, {'foo': 2}),
            types.Transition({'foo': 2}, 0, 0.0, 1.0, {'foo': 3}),
            types.Transition({'foo': 3}, 0, 1.0, 0.0, {'foo': 4}),
        )),
    dict(
        testcase_name='OneStepExtras',
        n_step=1,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (
                0,
                dm_env.transition(reward=0.0, observation=2),
                {
                    'state': 0
                },
            ),
            (
                0,
                dm_env.transition(reward=0.0, observation=3),
                {
                    'state': 1
                },
            ),
            (
                0,
                dm_env.termination(reward=1.0, observation=4),
                {
                    'state': 2
                },
            ),
        ),
        expected_transitions=(
            types.Transition(1, 0, 0.0, 1.0, 2, {'state': 0}),
            types.Transition(2, 0, 0.0, 1.0, 3, {'state': 1}),
            types.Transition(3, 0, 1.0, 0.0, 4, {'state': 2}),
        )),
    dict(
        testcase_name='OneStepExtrasZeroes',
        n_step=1,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (
                0,
                dm_env.transition(reward=0.0, observation=2),
                np.zeros(1),
            ),
            (
                0,
                dm_env.transition(reward=0.0, observation=3),
                np.zeros(1),
            ),
            (
                0,
                dm_env.termination(reward=1.0, observation=4),
                np.zeros(1),
            ),
        ),
        expected_transitions=(
            types.Transition(1, 0, 0.0, 1.0, 2, np.zeros(1)),
            types.Transition(2, 0, 0.0, 1.0, 3, np.zeros(1)),
            types.Transition(3, 0, 1.0, 0.0, 4, np.zeros(1)),
        )),
    dict(
        testcase_name='TwoStep',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=1.0, observation=2, discount=0.5)),
            (0, dm_env.transition(reward=1.0, observation=3, discount=0.5)),
            (0, dm_env.termination(reward=1.0, observation=4)),
        ),
        expected_transitions=(
            types.Transition(1, 0, 1.0, 0.50, 2),
            types.Transition(1, 0, 1.5, 0.25, 3),
            types.Transition(2, 0, 1.5, 0.00, 4),
            types.Transition(3, 0, 1.0, 0.00, 4),
        )),
    dict(
        testcase_name='TwoStepStructuredReward',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0,
             dm_env.transition(reward=(1.0, 2.0), observation=2, discount=0.5)),
            (0,
             dm_env.transition(reward=(1.0, 2.0), observation=3, discount=0.5)),
            (0, dm_env.termination(reward=(1.0, 2.0), observation=4)),
        ),
        expected_transitions=(
            types.Transition(1, 0, (1.0, 2.0), (0.50, 0.50), 2),
            types.Transition(1, 0, (1.5, 3.0), (0.25, 0.25), 3),
            types.Transition(2, 0, (1.5, 3.0), (0.00, 0.00), 4),
            types.Transition(3, 0, (1.0, 2.0), (0.00, 0.00), 4),
        )),
    dict(
        testcase_name='TwoStepNDArrayReward',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0,
             dm_env.transition(
                 reward=np.array((1.0, 2.0)), observation=2, discount=0.5)),
            (0,
             dm_env.transition(
                 reward=np.array((1.0, 2.0)), observation=3, discount=0.5)),
            (0, dm_env.termination(reward=np.array((1.0, 2.0)), observation=4)),
        ),
        expected_transitions=(
            types.Transition(1, 0, np.array((1.0, 2.0)), np.array((0.50, 0.50)),
                             2),
            types.Transition(1, 0, np.array((1.5, 3.0)), np.array((0.25, 0.25)),
                             3),
            types.Transition(2, 0, np.array((1.5, 3.0)), np.array((0.00, 0.00)),
                             4),
            types.Transition(3, 0, np.array((1.0, 2.0)), np.array((0.00, 0.00)),
                             4),
        )),
    dict(
        testcase_name='TwoStepStructuredDiscount',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0,
             dm_env.transition(
                 reward=1.0, observation=2, discount={
                     'a': 0.5,
                     'b': 0.1
                 })),
            (0,
             dm_env.transition(
                 reward=1.0, observation=3, discount={
                     'a': 0.5,
                     'b': 0.1
                 })),
            (0, dm_env.termination(reward=1.0,
                                   observation=4)._replace(discount={
                                       'a': 0.0,
                                       'b': 0.0
                                   })),
        ),
        expected_transitions=(
            types.Transition(1, 0, {
                'a': 1.0,
                'b': 1.0
            }, {
                'a': 0.50,
                'b': 0.10
            }, 2),
            types.Transition(1, 0, {
                'a': 1.5,
                'b': 1.1
            }, {
                'a': 0.25,
                'b': 0.01
            }, 3),
            types.Transition(2, 0, {
                'a': 1.5,
                'b': 1.1
            }, {
                'a': 0.00,
                'b': 0.00
            }, 4),
            types.Transition(3, 0, {
                'a': 1.0,
                'b': 1.0
            }, {
                'a': 0.00,
                'b': 0.00
            }, 4),
        )),
    dict(
        testcase_name='TwoStepNDArrayDiscount',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0,
             dm_env.transition(
                 reward=1.0, observation=2, discount=np.array((0.5, 0.1)))),
            (0,
             dm_env.transition(
                 reward=1.0, observation=3, discount=np.array((0.5, 0.1)))),
            (0, dm_env.termination(
                reward=1.0,
                observation=4)._replace(discount=np.array((0.0, 0.0)))),
        ),
        expected_transitions=(
            types.Transition(1, 0, np.array((1.0, 1.0)), np.array((0.50, 0.10)),
                             2),
            types.Transition(1, 0, np.array((1.5, 1.1)), np.array((0.25, 0.01)),
                             3),
            types.Transition(2, 0, np.array((1.5, 1.1)), np.array((0.00, 0.00)),
                             4),
            types.Transition(3, 0, np.array((1.0, 1.0)), np.array((0.00, 0.00)),
                             4),
        )),
    dict(
        testcase_name='TwoStepBroadcastedNDArrays',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0,
             dm_env.transition(
                 reward=np.array([[1.0, 2.0]]),
                 observation=2,
                 discount=np.array([[0.5], [0.1]]))),
            (0,
             dm_env.transition(
                 reward=np.array([[1.0, 2.0]]),
                 observation=3,
                 discount=np.array([[0.5], [0.1]]))),
            (0, dm_env.termination(
                reward=np.array([[1.0, 2.0]]),
                observation=4)._replace(discount=np.array([[0.0], [0.0]]))),
        ),
        expected_transitions=(
            types.Transition(1, 0, np.array([[1.0, 2.0], [1.0, 2.0]]),
                             np.array([[0.50], [0.10]]), 2),
            types.Transition(1, 0, np.array([[1.5, 3.0], [1.1, 2.2]]),
                             np.array([[0.25], [0.01]]), 3),
            types.Transition(2, 0, np.array([[1.5, 3.0], [1.1, 2.2]]),
                             np.array([[0.00], [0.00]]), 4),
            types.Transition(3, 0, np.array([[1.0, 2.0], [1.0, 2.0]]),
                             np.array([[0.00], [0.00]]), 4),
        )),
    dict(
        testcase_name='TwoStepStructuredBroadcastedNDArrays',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0,
             dm_env.transition(
                 reward={'a': np.array([[1.0, 2.0]])},
                 observation=2,
                 discount=np.array([[0.5], [0.1]]))),
            (0,
             dm_env.transition(
                 reward={'a': np.array([[1.0, 2.0]])},
                 observation=3,
                 discount=np.array([[0.5], [0.1]]))),
            (0,
             dm_env.termination(
                 reward={
                     'a': np.array([[1.0, 2.0]])
                 }, observation=4)._replace(discount=np.array([[0.0], [0.0]]))),
        ),
        expected_transitions=(
            types.Transition(1, 0, {'a': np.array([[1.0, 2.0], [1.0, 2.0]])},
                             {'a': np.array([[0.50], [0.10]])}, 2),
            types.Transition(1, 0, {'a': np.array([[1.5, 3.0], [1.1, 2.2]])},
                             {'a': np.array([[0.25], [0.01]])}, 3),
            types.Transition(2, 0, {'a': np.array([[1.5, 3.0], [1.1, 2.2]])},
                             {'a': np.array([[0.00], [0.00]])}, 4),
            types.Transition(3, 0, {'a': np.array([[1.0, 2.0], [1.0, 2.0]])},
                             {'a': np.array([[0.00], [0.00]])}, 4),
        )),
    dict(
        testcase_name='TwoStepWithExtras',
        n_step=2,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (
                0,
                dm_env.transition(reward=1.0, observation=2, discount=0.5),
                {
                    'state': 0
                },
            ),
            (
                0,
                dm_env.transition(reward=1.0, observation=3, discount=0.5),
                {
                    'state': 1
                },
            ),
            (
                0,
                dm_env.termination(reward=1.0, observation=4),
                {
                    'state': 2
                },
            ),
        ),
        expected_transitions=(
            types.Transition(1, 0, 1.0, 0.50, 2, {'state': 0}),
            types.Transition(1, 0, 1.5, 0.25, 3, {'state': 0}),
            types.Transition(2, 0, 1.5, 0.00, 4, {'state': 1}),
            types.Transition(3, 0, 1.0, 0.00, 4, {'state': 2}),
        )),
    dict(
        testcase_name='ThreeStepDiscounted',
        n_step=3,
        additional_discount=0.4,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=1.0, observation=2, discount=0.5)),
            (0, dm_env.transition(reward=1.0, observation=3, discount=0.5)),
            (0, dm_env.termination(reward=1.0, observation=4)),
        ),
        expected_transitions=(
            types.Transition(1, 0, 1.00, 0.5, 2),
            types.Transition(1, 0, 1.20, 0.1, 3),
            types.Transition(1, 0, 1.24, 0.0, 4),
            types.Transition(2, 0, 1.20, 0.0, 4),
            types.Transition(3, 0, 1.00, 0.0, 4),
        )),
    dict(
        testcase_name='ThreeStepVaryingReward',
        n_step=3,
        additional_discount=0.5,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.termination(reward=7.0, observation=5)),
        ),
        expected_transitions=(
            types.Transition(1, 0, 2, 1.00, 2),
            types.Transition(1, 0, 2 + 0.5 * 3, 0.50, 3),
            types.Transition(1, 0, 2 + 0.5 * 3 + 0.25 * 5, 0.25, 4),
            types.Transition(2, 0, 3 + 0.5 * 5 + 0.25 * 7, 0.00, 5),
            types.Transition(3, 0, 5 + 0.5 * 7, 0.00, 5),
            types.Transition(4, 0, 7, 0.00, 5),
        )),
    dict(
        testcase_name='SingleTransitionEpisode',
        n_step=4,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=((0, dm_env.termination(reward=1.0, observation=2)),),
        expected_transitions=(types.Transition(1, 0, 1.00, 0.0, 2),)),
    dict(
        testcase_name='EpisodeShorterThanN',
        n_step=4,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=1.0, observation=2)),
            (0, dm_env.termination(reward=1.0, observation=3)),
        ),
        expected_transitions=(
            types.Transition(1, 0, 1.00, 1.0, 2),
            types.Transition(1, 0, 2.00, 0.0, 3),
            types.Transition(2, 0, 1.00, 0.0, 3),
        )),
    dict(
        testcase_name='EpisodeEqualToN',
        n_step=3,
        additional_discount=1.0,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=1.0, observation=2)),
            (0, dm_env.termination(reward=1.0, observation=3)),
        ),
        expected_transitions=(
            types.Transition(1, 0, 1.00, 1.0, 2),
            types.Transition(1, 0, 2.00, 0.0, 3),
            types.Transition(2, 0, 1.00, 0.0, 3),
        )),
]

TEST_CASES_FOR_SEQUENCE_ADDER = [
    dict(
        testcase_name='PeriodOne',
        sequence_length=3,
        period=1,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.termination(reward=7.0, observation=5)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 0.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 0.0, False, ()),
                (5, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='PeriodTwo',
        sequence_length=3,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.termination(reward=7.0, observation=5)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 0.0, False, ()),
                (5, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='EarlyTerminationPeriodOne',
        sequence_length=3,
        period=1,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.termination(reward=3.0, observation=3)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 0.0, False, ()),
                (3, 0, 0.0, 0.0, False, ()),
            ],),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='EarlyTerminationPeriodTwo',
        sequence_length=3,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.termination(reward=3.0, observation=3)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 0.0, False, ()),
                (3, 0, 0.0, 0.0, False, ()),
            ],),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='EarlyTerminationPaddingPeriodOne',
        sequence_length=4,
        period=1,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.termination(reward=3.0, observation=3)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 0.0, False, ()),
                (3, 0, 0.0, 0.0, False, ()),
                (0, 0, 0.0, 0.0, False, ()),
            ],),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='EarlyTerminationPaddingPeriodTwo',
        sequence_length=4,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.termination(reward=3.0, observation=3)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 0.0, False, ()),
                (3, 0, 0.0, 0.0, False, ()),
                (0, 0, 0.0, 0.0, False, ()),
            ],),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='EarlyTerminationNoPadding',
        sequence_length=4,
        period=1,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.termination(reward=3.0, observation=3)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 0.0, False, ()),
                (3, 0, 0.0, 0.0, False, ()),
            ],),
        end_behavior=sequence_adder.EndBehavior.TRUNCATE,
    ),
    dict(
        testcase_name='LongEpisodePadding',
        sequence_length=3,
        period=3,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.transition(reward=7.0, observation=5)),
            (0, dm_env.transition(reward=9.0, observation=6)),
            (0, dm_env.transition(reward=11.0, observation=7)),
            (0, dm_env.termination(reward=13.0, observation=8)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 11.0, 1.0, False, ()),
            ],
            [
                (7, 0, 13.0, 0.0, False, ()),
                (8, 0, 0.0, 0.0, False, ()),
                (0, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.ZERO_PAD,
    ),
    dict(
        testcase_name='LongEpisodeNoPadding',
        sequence_length=3,
        period=3,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.transition(reward=7.0, observation=5)),
            (0, dm_env.transition(reward=9.0, observation=6)),
            (0, dm_env.transition(reward=11.0, observation=7)),
            (0, dm_env.termination(reward=13.0, observation=8)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 11.0, 1.0, False, ()),
            ],
            [
                (7, 0, 13.0, 0.0, False, ()),
                (8, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.TRUNCATE,
    ),
    dict(
        testcase_name='NonBreakingSequenceOnEpisodeReset',
        sequence_length=3,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.transition(reward=7.0, observation=5)),
            (0, dm_env.transition(reward=9.0, observation=6)),
            (0, dm_env.transition(reward=11.0, observation=7)),
            (0, dm_env.termination(reward=13.0, observation=8)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
            ],
            [
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 11.0, 1.0, False, ()),
                (7, 0, 13.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.CONTINUE,
        repeat_episode_times=1),
    dict(
        testcase_name='NonBreakingSequenceMultipleTerminatedEpisodes',
        sequence_length=3,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.transition(reward=7.0, observation=5)),
            (0, dm_env.transition(reward=9.0, observation=6)),
            (0, dm_env.termination(reward=13.0, observation=7)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
            ],
            [
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 13.0, 0.0, False, ()),
                (7, 0, 0.0, 0.0, False, ()),
            ],
            [
                (7, 0, 0.0, 0.0, False, ()),
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
            ],
            [
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
            ],
            [
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 13.0, 0.0, False, ()),
            ],
            [
                (6, 0, 13.0, 0.0, False, ()),
                (7, 0, 0.0, 0.0, False, ()),
                (1, 0, 2.0, 1.0, True, ()),
            ],
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
            ],
            [
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 13.0, 0.0, False, ()),
                (7, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.CONTINUE,
        repeat_episode_times=3),
    dict(
        testcase_name='NonBreakingSequenceMultipleTruncatedEpisodes',
        sequence_length=3,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.transition(reward=7.0, observation=5)),
            (0, dm_env.transition(reward=9.0, observation=6)),
            (0, dm_env.truncation(reward=13.0, observation=7)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
            ],
            [
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 13.0, 1.0, False, ()),
                (7, 0, 0.0, 0.0, False, ()),
            ],
            [
                (7, 0, 0.0, 0.0, False, ()),
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
            ],
            [
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
            ],
            [
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 13.0, 1.0, False, ()),
            ],
            [
                (6, 0, 13.0, 1.0, False, ()),
                (7, 0, 0.0, 0.0, False, ()),
                (1, 0, 2.0, 1.0, True, ()),
            ],
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 9.0, 1.0, False, ()),
            ],
            [
                (5, 0, 9.0, 1.0, False, ()),
                (6, 0, 13.0, 1.0, False, ()),
                (7, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.CONTINUE,
        repeat_episode_times=3),
    dict(
        testcase_name='EndBehavior_WRITE',
        sequence_length=3,
        period=2,
        first=dm_env.restart(1),
        steps=(
            (0, dm_env.transition(reward=2.0, observation=2)),
            (0, dm_env.transition(reward=3.0, observation=3)),
            (0, dm_env.transition(reward=5.0, observation=4)),
            (0, dm_env.transition(reward=7.0, observation=5)),
            (0, dm_env.termination(reward=8.0, observation=6)),
        ),
        expected_sequences=(
            # (observation, action, reward, discount, start_of_episode, extra)
            [
                (1, 0, 2.0, 1.0, True, ()),
                (2, 0, 3.0, 1.0, False, ()),
                (3, 0, 5.0, 1.0, False, ()),
            ],
            [
                (3, 0, 5.0, 1.0, False, ()),
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 8.0, 0.0, False, ()),
            ],
            [
                (4, 0, 7.0, 1.0, False, ()),
                (5, 0, 8.0, 0.0, False, ()),
                (6, 0, 0.0, 0.0, False, ()),
            ],
        ),
        end_behavior=sequence_adder.EndBehavior.WRITE,
    ),
]



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools to train a policy network with BC."""
from typing import Callable, Iterator

from acme import types
from acme.agents.jax.bc import learning
from acme.agents.jax.bc import losses
from acme.jax import networks as networks_lib
import jax
import optax


def train_with_bc(make_demonstrations: Callable[[int],
                                                Iterator[types.Transition]],
                  networks: networks_lib.FeedForwardNetwork,
                  loss: losses.Loss,
                  num_steps: int = 100000) -> networks_lib.Params:
  """Trains the given network with BC and returns the params.

  Args:
    make_demonstrations: A function (batch_size) -> iterator with demonstrations
      to be imitated.
    networks: Network taking (params, obs, is_training, key) as input
    loss: BC loss to use.
    num_steps: number of training steps

  Returns:
    The trained network params.
  """
  demonstration_iterator = make_demonstrations(256)

  learner = learning.BCLearner(
      network=networks, random_key=jax.random.PRNGKey(0), loss_fn=loss,
      demonstrations=demonstration_iterator, optimizer=optax.adam(1e-4),
      num_sgd_steps_per_step=1)

  # Train the agent
  for _ in range(num_steps):
    learner.step()

  return learner.get_variables(['policy'])[0]


def convert_to_bc_network(
    policy_network: networks_lib.FeedForwardNetwork
) -> networks_lib.FeedForwardNetwork:
  """Converts a policy_network from SAC/TD3/D4PG/.. into a BC policy network.

  Args:
    policy_network: FeedForwardNetwork taking the observation as input and
      returning action representation compatible with one of the BC losses.

  Returns:
    The BC policy network taking observation, is_training, key as input.
  """

  def apply(params, obs, is_training=False, key=None):
    del is_training, key
    return policy_network.apply(params, obs)

  return networks_lib.FeedForwardNetwork(policy_network.init, apply)


def convert_policy_value_to_bc_network(
    policy_value_network: networks_lib.FeedForwardNetwork
) -> networks_lib.FeedForwardNetwork:
  """Converts a network from e.g. PPO into a BC policy network.

  Args:
    policy_value_network: FeedForwardNetwork taking the observation as input.

  Returns:
    The BC policy network taking observation, is_training, key as input.
  """

  def apply(params, obs, is_training=False, key=None):
    del is_training, key
    actions, _ = policy_value_network.apply(params, obs)
    return actions

  return networks_lib.FeedForwardNetwork(policy_value_network.init, apply)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme.tf.variable_utils."""

import threading

from acme.testing import fakes
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils
import sonnet as snt
import tensorflow as tf


_MLP_LAYERS = [50, 30]
_INPUT_SIZE = 28
_BATCH_SIZE = 8
_UPDATE_PERIOD = 2


class VariableClientTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()

    # Create two instances of the same model.
    self._actor_model = snt.nets.MLP(_MLP_LAYERS)
    self._learner_model = snt.nets.MLP(_MLP_LAYERS)

    # Create variables first.
    input_spec = tf.TensorSpec(shape=(_INPUT_SIZE,), dtype=tf.float32)
    tf2_utils.create_variables(self._actor_model, [input_spec])
    tf2_utils.create_variables(self._learner_model, [input_spec])

  def test_update_and_wait(self):
    # Create a variable source (emulating the learner).
    np_learner_variables = tf2_utils.to_numpy(self._learner_model.variables)
    variable_source = fakes.VariableSource(np_learner_variables)

    # Create a variable client (emulating the actor).
    variable_client = tf2_variable_utils.VariableClient(
        variable_source, {'policy': self._actor_model.variables})

    # Create some random batch of test input:
    x = tf.random.normal(shape=(_BATCH_SIZE, _INPUT_SIZE))

    # Before copying variables, the models have different outputs.
    self.assertNotAllClose(self._actor_model(x), self._learner_model(x))

    # Update the variable client.
    variable_client.update_and_wait()

    # After copying variables (by updating the client), the models are the same.
    self.assertAllClose(self._actor_model(x), self._learner_model(x))

  def test_update(self):
    # Create a barrier to be shared between the test body and the variable
    # source. The barrier will block until, in this case, two threads call
    # wait(). Note that the (fake) variable source will call it within its
    # get_variables() call.
    barrier = threading.Barrier(2)

    # Create a variable source (emulating the learner).
    np_learner_variables = tf2_utils.to_numpy(self._learner_model.variables)
    variable_source = fakes.VariableSource(np_learner_variables, barrier)

    # Create a variable client (emulating the actor).
    variable_client = tf2_variable_utils.VariableClient(
        variable_source, {'policy': self._actor_model.variables},
        update_period=_UPDATE_PERIOD)

    # Create some random batch of test input:
    x = tf.random.normal(shape=(_BATCH_SIZE, _INPUT_SIZE))

    # Create variables by doing the computation once.
    learner_output = self._learner_model(x)
    actor_output = self._actor_model(x)
    del learner_output, actor_output

    for _ in range(_UPDATE_PERIOD):
      # Before the update period is reached, the models have different outputs.
      self.assertNotAllClose(self._actor_model.variables,
                             self._learner_model.variables)

      # Before the update period is reached, the variable client should not make
      # any requests for variables.
      self.assertIsNone(variable_client._future)

      variable_client.update()

    # Make sure the last call created a request for variables and reset the
    # internal call counter.
    self.assertIsNotNone(variable_client._future)
    self.assertEqual(variable_client._call_counter, 0)
    future = variable_client._future

    for _ in range(_UPDATE_PERIOD):
      # Before the barrier allows the variables to be released, the models have
      # different outputs.
      self.assertNotAllClose(self._actor_model.variables,
                             self._learner_model.variables)

      variable_client.update()

      # Make sure no new requests are made.
      self.assertEqual(variable_client._future, future)

    # Calling wait() on the barrier will now allow the variables to be copied
    # over from source to client.
    barrier.wait()

    # Update once more to ensure the variables are copied over.
    while variable_client._future is not None:
      variable_client.update()

    # After a number of update calls, the variables should be the same.
    self.assertAllClose(self._actor_model.variables,
                        self._learner_model.variables)


if __name__ == '__main__':
  tf.test.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for JAX."""

import functools
import itertools
import queue
import threading
from typing import Callable, Generator, Iterable, NamedTuple, Optional, Sequence, Tuple, TypeVar

from absl import logging
from acme import types
from acme.jax.types import TrainingMetrics, TrainingStepOutput  # pylint: disable=g-multiple-import
import jax
import jax.numpy as jnp
import numpy as np
import tree

F = TypeVar('F', bound=Callable)
N = TypeVar('N', bound=types.NestedArray)
T = TypeVar('T')


def add_batch_dim(values: types.Nest) -> types.NestedArray:
  return jax.tree_map(lambda x: jnp.expand_dims(x, axis=0), values)


def _flatten(x: jnp.ndarray, num_batch_dims: int) -> jnp.ndarray:
  """Flattens the input, preserving the first ``num_batch_dims`` dimensions.

  If the input has fewer than ``num_batch_dims`` dimensions, it is returned
  unchanged.
  If the input has exactly ``num_batch_dims`` dimensions, an extra dimension
  is added. This is needed to handle batched scalars.

  Arguments:
    x: the input array to flatten.
    num_batch_dims: number of dimensions to preserve.

  Returns:
    flattened input.
  """
  # TODO(b/173492429): consider throwing an error instead.
  if x.ndim < num_batch_dims:
    return x
  return jnp.reshape(x, list(x.shape[:num_batch_dims]) + [-1])


def batch_concat(
    values: types.NestedArray,
    num_batch_dims: int = 1,
) -> jnp.ndarray:
  """Flatten and concatenate nested array structure, keeping batch dims."""
  flatten_fn = lambda x: _flatten(x, num_batch_dims)
  flat_leaves = tree.map_structure(flatten_fn, values)
  return jnp.concatenate(tree.flatten(flat_leaves), axis=-1)


def zeros_like(nest: types.Nest, dtype=None) -> types.NestedArray:
  return jax.tree_map(lambda x: jnp.zeros(x.shape, dtype or x.dtype), nest)


def ones_like(nest: types.Nest, dtype=None) -> types.NestedArray:
  return jax.tree_map(lambda x: jnp.ones(x.shape, dtype or x.dtype), nest)


def squeeze_batch_dim(nest: types.Nest) -> types.NestedArray:
  return jax.tree_map(lambda x: jnp.squeeze(x, axis=0), nest)


def to_numpy_squeeze(values: types.Nest) -> types.NestedArray:
  """Converts to numpy and squeezes out dummy batch dimension."""
  return jax.tree_map(lambda x: np.asarray(x).squeeze(axis=0), values)


def to_numpy(values: types.Nest) -> types.NestedArray:
  return jax.tree_map(np.asarray, values)


def fetch_devicearray(values: types.Nest) -> types.Nest:
  """Fetches and converts any DeviceArrays to np.ndarrays."""
  return tree.map_structure(_fetch_devicearray, values)


def _fetch_devicearray(x):
  if isinstance(x, jax.xla.DeviceArray):
    return np.asarray(x)
  return x


def batch_to_sequence(values: types.Nest) -> types.NestedArray:
  return jax.tree_map(
      lambda x: jnp.transpose(x, axes=(1, 0, *range(2, len(x.shape)))), values)


def tile_array(array: jnp.ndarray, multiple: int) -> jnp.ndarray:
  """Tiles `multiple` copies of `array` along a new leading axis."""
  return jnp.stack([array] * multiple)


def tile_nested(inputs: types.Nest, multiple: int) -> types.Nest:
  """Tiles tensors in a nested structure along a new leading axis."""
  tile = functools.partial(tile_array, multiple=multiple)
  return jax.tree_map(tile, inputs)


def prefetch(iterable: Iterable[T],
             buffer_size: int = 5,
             device=None) -> Generator[T, None, None]:
  """Performs prefetching of elements from an iterable in a separate thread.

  Args:
    iterable: A python iterable. This is used to build the python prefetcher.
      Note that each iterable should only be passed to this function once as
      iterables aren't thread safe
    buffer_size (int): Number of elements to keep in the prefetch buffer.
    device: The device to prefetch the elements to. If none then the elements
      are left on the CPU. The device should be of the type returned by
      `jax.devices()`.

  Yields:
    Prefetched elements from the original iterable.
  Raises:
    ValueError if the buffer_size <= 1.
    Any error thrown by the iterable_function. Note this is not raised inside
      the producer, but after it finishes executing.
  """

  if buffer_size <= 1:
    raise ValueError('the buffer_size should be > 1')
  buffer = queue.Queue(maxsize=(buffer_size - 1))
  producer_error = []
  end = object()

  def producer():
    """Enqueues items from `iterable` on a given thread."""
    try:
      # Build a new iterable for each thread. This is crucial if working with
      # tensorflow datasets because tf.Graph objects are thread local.
      for item in iterable:
        if device:
          item = jax.device_put(item, device)
        buffer.put(item)
    except Exception as e:  # pylint: disable=broad-except
      logging.exception('Error in producer thread for %s', iterable)
      producer_error.append(e)
    finally:
      buffer.put(end)

  # Start the producer thread.
  threading.Thread(target=producer, daemon=True).start()

  # Consume from the buffer.
  while True:
    value = buffer.get()
    if value is end:
      break
    yield value

  if producer_error:
    raise producer_error[0]


class PrefetchingSplit(NamedTuple):
  host: types.NestedArray
  device: types.NestedArray


_SplitFunction = Callable[[types.NestedArray], PrefetchingSplit]


def sharded_prefetch(
    iterable: Iterable[types.NestedArray],
    buffer_size: int = 5,
    num_threads: int = 1,
    split_fn: Optional[_SplitFunction] = None,
    devices: Optional[Sequence[jax.xla.Device]] = None,
) -> Generator[types.NestedArray, None, None]:
  """Performs sharded prefetching from an iterable in separate threads.

  Elements from the resulting generator are intended to be used in a jax.pmap
  call. Every element is a sharded prefetched array with an additional replica
  dimension and corresponds to jax.local_device_count() elements from the
  original iterable.

  Args:
    iterable: A python iterable. This is used to build the python prefetcher.
      Note that each iterable should only be passed to this function once as
      iterables aren't thread safe.
    buffer_size (int): Number of elements to keep in the prefetch buffer.
    num_threads (int): Number of threads.
    split_fn: Optional function applied to every element from the iterable to
      split the parts of it that will be kept in the host and the parts that
      will sent to the device.
    devices: Devices used for prefecthing. Optional, jax.local_devices() by
      default.

  Yields:
    Prefetched elements from the original iterable with additional replica
    dimension.
  Raises:
    ValueError if the buffer_size <= 1.
    Any error thrown by the iterable_function. Note this is not raised inside
      the producer, but after it finishes executing.
  """

  devices = devices or jax.local_devices()

  if buffer_size <= 1:
    raise ValueError('the buffer_size should be > 1')
  buffer = queue.Queue(maxsize=(buffer_size - 1))
  producer_error = []
  end = object()

  def producer():
    """Enqueues batched items from `iterable` on a given thread."""
    try:
      # Build a new iterable for each thread. This is crucial if working with
      # tensorflow datasets because tf.Graph objects are thread local.
      it = iter(iterable)
      while True:
        items = itertools.islice(it, len(devices))
        if not items:
          break
        if split_fn is None:
          buffer.put(jax.device_put_sharded(tuple(items), devices))
        else:
          # ((host: x1, device: y1), ..., (host: xN, device: yN)).
          items_split = (split_fn(item) for item in items)
          # (host: (x1, ..., xN), device: (y1, ..., yN)).
          split = tree.map_structure_up_to(
              PrefetchingSplit(None, None), lambda *x: x, *items_split)

          buffer.put(
              PrefetchingSplit(
                  host=np.stack(split.host),
                  device=jax.device_put_sharded(split.device, devices)))
    except Exception as e:  # pylint: disable=broad-except
      logging.exception('Error in producer thread for %s', iterable)
      producer_error.append(e)
    finally:
      buffer.put(end)

  # Start producer threads.
  for _ in range(num_threads):
    threading.Thread(target=producer, daemon=True).start()

  # Consume from the buffer.
  while True:
    value = buffer.get()
    if value is end:
      break
    yield value

  if producer_error:
    raise producer_error[0]


def replicate_in_all_devices(nest: N,
                             devices: Optional[Sequence[jax.xla.Device]] = None
                            ) -> N:
  """Replicate array nest in all available devices."""
  devices = devices or jax.local_devices()
  return jax.device_put_sharded([nest] * len(devices), devices)


def get_from_first_device(nest: N, as_numpy: bool = True) -> N:
  """Gets the first array of a nest of `jax.pxla.ShardedDeviceArray`s.

  Args:
    nest: A nest of `jax.pxla.ShardedDeviceArray`s.
    as_numpy: If `True` then each `DeviceArray` that is retrieved is transformed
      (and copied if not on the host machine) into a `np.ndarray`.

  Returns:
    The first array of a nest of `jax.pxla.ShardedDeviceArray`s. Note that if
    `as_numpy=False` then the array will be a `DeviceArray` (which will live on
    the same device as the sharded device array). If `as_numpy=True` then the
    array will be copied to the host machine and converted into a `np.ndarray`.
  """

  def _check_type_and_slice(x):
    if not isinstance(x, jax.pxla.ShardedDeviceArray):
      raise ValueError('get_from_first_device should only be used with '
                       f'{jax.pxla.ShardedDeviceArray}, passed {type(x)}.')
    return x[0]

  zeroth_nest = jax.tree_map(_check_type_and_slice, nest)

  return jax.device_get(zeroth_nest) if as_numpy else zeroth_nest


def mapreduce(
    f: F,
    reduce_fn: Optional[Callable[[jnp.DeviceArray], jnp.DeviceArray]] = None,
    **vmap_kwargs,
) -> F:
  """A simple decorator that transforms `f` into (`reduce_fn` o vmap o f).

  By default, we vmap over axis 0, and the `reduce_fn` is jnp.mean over axis 0.
  Note that the call signature of `f` is invariant under this transformation.

  If, for example, f has shape signature [H, W] -> [N], then mapreduce(f)
  (with the default arguments) will have shape signature [B, H, W] -> [N].

  Args:
    f: A pure function over examples.
    reduce_fn: A pure function that reduces DeviceArrays -> DeviceArrays.
    **vmap_kwargs: Keyword arguments to forward to `jax.vmap`.

  Returns:
    g: A pure function over batches of examples.
  """

  if reduce_fn is None:
    reduce_fn = lambda x: jnp.mean(x, axis=0)

  vmapped_f = jax.vmap(f, **vmap_kwargs)

  def g(*args, **kwargs):
    return jax.tree_map(reduce_fn, vmapped_f(*args, **kwargs))

  return g


_TrainingState = TypeVar('_TrainingState')
_TrainingData = TypeVar('_TrainingData')
_TrainingAux = TypeVar('_TrainingAux')


# TODO(b/192806089): migrate all callers to process_many_batches and remove this
# method.
def process_multiple_batches(
    process_one_batch: Callable[[_TrainingState, _TrainingData],
                                Tuple[_TrainingState, _TrainingAux]],
    num_batches: int,
    postprocess_aux: Optional[Callable[[_TrainingAux], _TrainingAux]] = None
) -> Callable[[_TrainingState, _TrainingData], Tuple[_TrainingState,
                                                     _TrainingAux]]:
  """Makes 'process_one_batch' process multiple batches at once.

  Args:
    process_one_batch: a function that takes 'state' and 'data', and returns
      'new_state' and 'aux' (for example 'metrics').
    num_batches: how many batches to process at once
    postprocess_aux: how to merge the extra information, defaults to taking
      the mean.
  Returns:
    A function with the same interface as 'process_one_batch' which processes
    multiple batches at once.
  """
  assert num_batches >= 1
  if num_batches == 1:
    if not postprocess_aux:
      return process_one_batch
    def _process_one_batch(state, data):
      state, aux = process_one_batch(state, data)
      return state, postprocess_aux(aux)
    return _process_one_batch

  if postprocess_aux is None:
    postprocess_aux = lambda x: jax.tree_map(jnp.mean, x)

  def _process_multiple_batches(state, data):
    data = jax.tree_map(
        lambda a: jnp.reshape(a, (num_batches, -1, *a.shape[1:])), data)

    state, aux = jax.lax.scan(
        process_one_batch, state, data, length=num_batches)
    return state, postprocess_aux(aux)

  return _process_multiple_batches


def process_many_batches(
    process_one_batch: Callable[[_TrainingState, _TrainingData],
                                TrainingStepOutput[_TrainingState]],
    num_batches: int,
    postprocess_aux: Optional[Callable[[TrainingMetrics],
                                       TrainingMetrics]] = None
) -> Callable[[_TrainingState, _TrainingData],
              TrainingStepOutput[_TrainingState]]:
  """The version of 'process_multiple_batches' with stronger typing."""

  def _process_one_batch(
      state: _TrainingState,
      data: _TrainingData) -> Tuple[_TrainingState, TrainingMetrics]:
    result = process_one_batch(state, data)
    return result.state, result.metrics

  func = process_multiple_batches(_process_one_batch, num_batches,
                                  postprocess_aux)

  def _process_many_batches(
      state: _TrainingState,
      data: _TrainingData) -> TrainingStepOutput[_TrainingState]:
    state, aux = func(state, data)
    return TrainingStepOutput(state, aux)

  return _process_many_batches


def weighted_softmax(x: jnp.ndarray, weights: jnp.ndarray, axis: int = 0):
  x = x - jnp.max(x, axis=axis)
  return weights * jnp.exp(x) / jnp.sum(weights * jnp.exp(x),
                                        axis=axis, keepdims=True)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Networks definitions for the BC agent.

AIRL network architecture follows https://arxiv.org/pdf/1710.11248.pdf.
"""
import dataclasses
import functools
from typing import Any, Callable, Generic, Iterable, Optional

from acme import specs
from acme import types
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.jax.imitation_learning_types import DirectRLNetworks
import haiku as hk
import jax
from jax import numpy as jnp
import numpy as np

# Function from discriminator logit to imitation reward.
ImitationRewardFn = Callable[[networks_lib.Logits], jnp.ndarray]
State = networks_lib.Params


@dataclasses.dataclass
class AILNetworks(Generic[DirectRLNetworks]):
  """AIL networks data class.

  Attributes:
    discriminator_network: Networks which takes as input:
      (observations, actions, next_observations, direct_rl_params)
      to return the logit of the discriminator.
      If the discriminator does not need direct_rl_params you can pass ().
    imitation_reward_fn: Function from logit of the discriminator to imitation
      reward.
    direct_rl_networks: Networks of the direct RL algorithm.
  """
  discriminator_network: networks_lib.FeedForwardNetwork
  imitation_reward_fn: ImitationRewardFn
  direct_rl_networks: DirectRLNetworks


def compute_ail_reward(discriminator_params: networks_lib.Params,
                       discriminator_state: State,
                       policy_params: Optional[networks_lib.Params],
                       transitions: types.Transition,
                       networks: AILNetworks) -> jnp.ndarray:
  """Computes the AIL reward for a given transition.

  Args:
    discriminator_params: Parameters of the discriminator network.
    discriminator_state: State of the discriminator network.
    policy_params: Parameters of the direct RL policy.
    transitions: Transitions to compute the reward for.
    networks: AIL networks.

  Returns:
    The rewards as an ndarray.
  """
  logits, _ = networks.discriminator_network.apply(
      discriminator_params,
      policy_params,
      discriminator_state,
      transitions,
      is_training=False,
      rng=None)
  return networks.imitation_reward_fn(logits)


class SpectralNormalizedLinear(hk.Module):
  """SpectralNormalizedLinear module.

  This is a Linear layer with a upper-bounded Lipschitz. It is used in iResNet.

  Reference:
    Behrmann et al. Invertible Residual Networks. ICML 2019.
    https://arxiv.org/pdf/1811.00995.pdf
  """

  def __init__(
      self,
      output_size: int,
      lipschitz_coeff: float,
      with_bias: bool = True,
      w_init: Optional[hk.initializers.Initializer] = None,
      b_init: Optional[hk.initializers.Initializer] = None,
      name: Optional[str] = None,
  ):
    """Constructs the SpectralNormalizedLinear module.

    Args:
      output_size: Output dimensionality.
      lipschitz_coeff: Spectral normalization coefficient.
      with_bias: Whether to add a bias to the output.
      w_init: Optional initializer for weights. By default, uses random values
        from truncated normal, with stddev ``1 / sqrt(fan_in)``. See
        https://arxiv.org/abs/1502.03167v3.
      b_init: Optional initializer for bias. By default, zero.
      name: Name of the module.
    """
    super().__init__(name=name)
    self.input_size = None
    self.output_size = output_size
    self.with_bias = with_bias
    self.w_init = w_init
    self.b_init = b_init or jnp.zeros
    self.lipschitz_coeff = lipschitz_coeff
    self.num_iterations = 100
    self.eps = 1e-6

  def get_normalized_weights(self,
                             weights: jnp.ndarray,
                             renormalize: bool = False) -> jnp.ndarray:

    def _l2_normalize(x, axis=None, eps=1e-12):
      return x * jax.lax.rsqrt((x * x).sum(axis=axis, keepdims=True) + eps)

    output_size = self.output_size
    dtype = weights.dtype
    assert output_size == weights.shape[-1]
    sigma = hk.get_state('sigma', (), init=jnp.ones)
    if renormalize:
      # Power iterations to compute spectral norm V*W*U^T.
      u = hk.get_state(
          'u', (1, output_size), dtype, init=hk.initializers.RandomNormal())
      for _ in range(self.num_iterations):
        v = _l2_normalize(jnp.matmul(u, weights.transpose()), eps=self.eps)
        u = _l2_normalize(jnp.matmul(v, weights), eps=self.eps)
      u = jax.lax.stop_gradient(u)
      v = jax.lax.stop_gradient(v)
      sigma = jnp.matmul(jnp.matmul(v, weights), jnp.transpose(u))[0, 0]
      hk.set_state('u', u)
      hk.set_state('v', v)
      hk.set_state('sigma', sigma)
    factor = jnp.maximum(1, sigma / self.lipschitz_coeff)
    return weights / factor

  def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
    """Computes a linear transform of the input."""
    if not inputs.shape:
      raise ValueError('Input must not be scalar.')

    input_size = self.input_size = inputs.shape[-1]
    output_size = self.output_size
    dtype = inputs.dtype

    w_init = self.w_init
    if w_init is None:
      stddev = 1. / np.sqrt(self.input_size)
      w_init = hk.initializers.TruncatedNormal(stddev=stddev)
    w = hk.get_parameter('w', [input_size, output_size], dtype, init=w_init)
    w = self.get_normalized_weights(w, renormalize=True)

    out = jnp.dot(inputs, w)

    if self.with_bias:
      b = hk.get_parameter('b', [self.output_size], dtype, init=self.b_init)
      b = jnp.broadcast_to(b, out.shape)
      out = out + b

    return out


class DiscriminatorMLP(hk.Module):
  """A multi-layer perceptron module."""

  def __init__(
      self,
      hidden_layer_sizes: Iterable[int],
      w_init: Optional[hk.initializers.Initializer] = None,
      b_init: Optional[hk.initializers.Initializer] = None,
      with_bias: bool = True,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.relu,
      input_dropout_rate: float = 0.,
      hidden_dropout_rate: float = 0.,
      spectral_normalization_lipschitz_coeff: Optional[float] = None,
      name: Optional[str] = None
  ):
    """Constructs an MLP.

    Args:
      hidden_layer_sizes: Hiddent layer sizes.
      w_init: Initializer for :class:`~haiku.Linear` weights.
      b_init: Initializer for :class:`~haiku.Linear` bias. Must be ``None`` if
        ``with_bias=False``.
      with_bias: Whether or not to apply a bias in each layer.
      activation: Activation function to apply between :class:`~haiku.Linear`
        layers. Defaults to ReLU.
      input_dropout_rate: Dropout on the input.
      hidden_dropout_rate: Dropout on the hidden layer outputs.
      spectral_normalization_lipschitz_coeff: If not None, the network will have
        spectral normalization with the given constant.
      name: Optional name for this module.

    Raises:
      ValueError: If ``with_bias`` is ``False`` and ``b_init`` is not ``None``.
    """
    if not with_bias and b_init is not None:
      raise ValueError('When with_bias=False b_init must not be set.')

    super().__init__(name=name)
    self._activation = activation
    self._input_dropout_rate = input_dropout_rate
    self._hidden_dropout_rate = hidden_dropout_rate
    layer_sizes = list(hidden_layer_sizes) + [1]

    if spectral_normalization_lipschitz_coeff is not None:
      layer_lipschitz_coeff = np.power(spectral_normalization_lipschitz_coeff,
                                       1. / len(layer_sizes))
      layer_module = functools.partial(
          SpectralNormalizedLinear,
          lipschitz_coeff=layer_lipschitz_coeff,
          w_init=w_init,
          b_init=b_init,
          with_bias=with_bias)
    else:
      layer_module = functools.partial(
          hk.Linear,
          w_init=w_init,
          b_init=b_init,
          with_bias=with_bias)

    layers = []
    for index, output_size in enumerate(layer_sizes):
      layers.append(
          layer_module(output_size=output_size, name=f'linear_{index}'))
    self._layers = tuple(layers)

  def __call__(
      self,
      inputs: jnp.ndarray,
      is_training: bool,
      rng: Optional[networks_lib.PRNGKey],
  ) -> networks_lib.Logits:
    rng = hk.PRNGSequence(rng) if rng is not None else None

    out = inputs
    for i, layer in enumerate(self._layers):
      if is_training:
        dropout_rate = (
            self._input_dropout_rate if i == 0 else self._hidden_dropout_rate)
        out = hk.dropout(next(rng), dropout_rate, out)
      out = layer(out)
      if i < len(self._layers) - 1:
        out = self._activation(out)

    return out


class DiscriminatorModule(hk.Module):
  """Discriminator module that concatenates its inputs."""

  def __init__(self,
               environment_spec: specs.EnvironmentSpec,
               use_action: bool,
               use_next_obs: bool,
               network_core: Callable[..., Any],
               observation_embedding: Callable[[networks_lib.Observation],
                                               jnp.ndarray] = lambda x: x,
               name='discriminator'):
    super().__init__(name=name)
    self._use_action = use_action
    self._environment_spec = environment_spec
    self._use_next_obs = use_next_obs
    self._network_core = network_core
    self._observation_embedding = observation_embedding

  def __call__(self, observations: networks_lib.Observation,
               actions: networks_lib.Action,
               next_observations: networks_lib.Observation, is_training: bool,
               rng: networks_lib.PRNGKey) -> networks_lib.Logits:
    observations = self._observation_embedding(observations)
    if self._use_next_obs:
      next_observations = self._observation_embedding(next_observations)
      data = jnp.concatenate([observations, next_observations], axis=-1)
    else:
      data = observations
    if self._use_action:
      action_spec = self._environment_spec.actions
      if isinstance(action_spec, specs.DiscreteArray):
        actions = jax.nn.one_hot(actions,
                                 action_spec.num_values)
      data = jnp.concatenate([data, actions], axis=-1)
    output = self._network_core(data, is_training, rng)
    output = jnp.squeeze(output, axis=-1)
    return output


class AIRLModule(hk.Module):
  """AIRL Module."""

  def __init__(self,
               environment_spec: specs.EnvironmentSpec,
               use_action: bool,
               use_next_obs: bool,
               discount: float,
               g_core: Callable[..., Any],
               h_core: Callable[..., Any],
               observation_embedding: Callable[[networks_lib.Observation],
                                               jnp.ndarray] = lambda x: x,
               name='airl'):
    super().__init__(name=name)
    self._environment_spec = environment_spec
    self._use_action = use_action
    self._use_next_obs = use_next_obs
    self._discount = discount
    self._g_core = g_core
    self._h_core = h_core
    self._observation_embedding = observation_embedding

  def __call__(self, observations: networks_lib.Observation,
               actions: networks_lib.Action,
               next_observations: networks_lib.Observation,
               is_training: bool,
               rng: networks_lib.PRNGKey) -> networks_lib.Logits:
    g_output = DiscriminatorModule(
        environment_spec=self._environment_spec,
        use_action=self._use_action,
        use_next_obs=self._use_next_obs,
        network_core=self._g_core,
        observation_embedding=self._observation_embedding,
        name='airl_g')(observations, actions, next_observations, is_training,
                       rng)
    h_module = DiscriminatorModule(
        environment_spec=self._environment_spec,
        use_action=False,
        use_next_obs=False,
        network_core=self._h_core,
        observation_embedding=self._observation_embedding,
        name='airl_h')
    return (g_output + self._discount * h_module(next_observations, (),
                                                 (), is_training, rng) -
            h_module(observations, (), (), is_training, rng))


# TODO(eorsini): Manipulate FeedForwardNetworks instead of transforms to
# increase compatibility with Flax.
def make_discriminator(
    environment_spec: specs.EnvironmentSpec,
    discriminator_transformed: hk.TransformedWithState,
    logpi_fn: Optional[Callable[
        [networks_lib.Params, networks_lib.Observation, networks_lib.Action],
        jnp.ndarray]] = None
) -> networks_lib.FeedForwardNetwork:
  """Creates the discriminator network.

  Args:
    environment_spec: Environment spec
    discriminator_transformed: Haiku transformed of the discriminator.
    logpi_fn: If the policy logpi function is provided, its output will be
      removed from the discriminator logit.

  Returns:
    The network.
  """

  def apply_fn(params: hk.Params,
               policy_params: networks_lib.Params,
               state: hk.State,
               transitions: types.Transition,
               is_training: bool,
               rng: networks_lib.PRNGKey) -> networks_lib.Logits:
    output, state = discriminator_transformed.apply(
        params, state, transitions.observation, transitions.action,
        transitions.next_observation, is_training, rng)
    if logpi_fn is not None:
      logpi = logpi_fn(policy_params, transitions.observation,
                       transitions.action)

      # Quick Maths:
      # D = exp(output)/(exp(output) + pi(a|s))
      # logit(D) = log(D/(1-D)) = log(exp(output)/pi(a|s))
      # logit(D) = output - logpi
      return output - logpi, state
    return output, state

  dummy_obs = utils.zeros_like(environment_spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)
  dummy_actions = utils.zeros_like(environment_spec.actions)
  dummy_actions = utils.add_batch_dim(dummy_actions)

  return networks_lib.FeedForwardNetwork(
      # pylint: disable=g-long-lambda
      init=lambda rng: discriminator_transformed.init(
          rng, dummy_obs, dummy_actions, dummy_obs, False, rng),
      apply=apply_fn)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration test for the distributed agent."""

from typing import Sequence

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.tf import svg0_prior
from acme.testing import fakes
from acme.tf import networks
from acme.tf import utils as tf2_utils
import launchpad as lp
import numpy as np
import sonnet as snt


def make_networks(
    action_spec: specs.BoundedArray,
    policy_layer_sizes: Sequence[int] = (10, 10),
    critic_layer_sizes: Sequence[int] = (10, 10),
):
  """Simple networks for testing.."""

  # Get total number of action dimensions from action spec.
  num_dimensions = np.prod(action_spec.shape, dtype=int)

  policy_network = snt.Sequential([
      tf2_utils.batch_concat,
      networks.LayerNormMLP(policy_layer_sizes, activate_final=True),
      networks.MultivariateNormalDiagHead(
          num_dimensions,
          tanh_mean=True,
          min_scale=0.3,
          init_scale=0.7,
          fixed_scale=False,
          use_tfd_independent=False)
  ])
  # The multiplexer concatenates the (maybe transformed) observations/actions.
  multiplexer = networks.CriticMultiplexer()
  critic_network = snt.Sequential([
      multiplexer,
      networks.LayerNormMLP(critic_layer_sizes, activate_final=True),
      networks.NearZeroInitializedLinear(1),
  ])

  return {
      'policy': policy_network,
      'critic': critic_network,
  }


class DistributedAgentTest(absltest.TestCase):
  """Simple integration/smoke test for the distributed agent."""

  def test_control_suite(self):
    """Tests that the agent can run on the control suite without crashing."""

    agent = svg0_prior.DistributedSVG0(
        environment_factory=lambda x: fakes.ContinuousEnvironment(),
        network_factory=make_networks,
        num_actors=2,
        batch_size=32,
        min_replay_size=32,
        max_replay_size=1000,
    )
    program = agent.build()

    (learner_node,) = program.groups['learner']
    learner_node.disable_run()

    lp.launch(program, launch_type='test_mt')

    learner: acme.Learner = learner_node.create_handle().dereference()

    for _ in range(5):
      learner.step()


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Install script for setuptools."""

import datetime
from importlib import util as import_util
import os
import sys

from setuptools import find_packages
from setuptools import setup
import setuptools.command.build_py
import setuptools.command.develop

spec = import_util.spec_from_file_location('_metadata', 'acme/_metadata.py')
_metadata = import_util.module_from_spec(spec)
spec.loader.exec_module(_metadata)

# TODO(b/184148890): Add a release flag


# Any particular version of reverb needs to be pinned against a particular
# version of TF due to how it is built. While the versions below should be the
# most recent stable versions of each library we'll be explicit just make make
# sure this constraint is upheld.
tensorflow = [
    'dm-reverb==0.6.1',
    'keras==2.7.0',
    'tensorflow-datasets==4.4.0',
    'tensorflow-estimator==2.7.0',
    'tensorflow==2.7.0',
    'tensorflow_probability==0.15.0',
]

core_requirements = [
    'absl-py',
    'dm-env',
    'dm-launchpad==0.4.1',
    'dm-tree',
    'numpy',
    'pillow',
    'typing-extensions',
]

jax_requirements = [
    'chex',
    'jax',
    'jaxlib',
    'dm-haiku',
    'flax',
    'optax',
    'rlax',
] + tensorflow

tf_requirements = [
    'dm-sonnet',
    'trfl',
] + tensorflow

testing_requirements = [
    'pytype==2021.8.11',  # TODO(b/206926677): update to new version.
    'pytest-xdist',
]

envs_requirements = [
    'atari-py',
    'bsuite',
    'dm-control==0.0.364896371',  # TODO(b/206925413): update to new version.
    'gym',
    'gym[atari]',
    'tensorflow_datasets',
]


def generate_requirements_file(path=None):
  """Generates requirements.txt file with the Acme's dependencies.

  It is used by Launchpad GCP runtime to generate Acme requirements to be
  installed inside the docker image. Acme itself is not installed from pypi,
  but instead sources are copied over to reflect any local changes made to
  the codebase.

  Args:
    path: path to the requirements.txt file to generate.
  """
  if not path:
    path = os.path.join(os.path.dirname(__file__), 'acme/requirements.txt')
  with open(path, 'w') as f:
    for package in set(core_requirements + jax_requirements + tf_requirements +
                       envs_requirements):
      f.write(f'{package}\n')


long_description = """Acme is a library of reinforcement learning (RL) agents
and agent building blocks. Acme strives to expose simple, efficient,
and readable agents, that serve both as reference implementations of popular
algorithms and as strong baselines, while still providing enough flexibility
to do novel research. The design of Acme also attempts to provide multiple
points of entry to the RL problem at differing levels of complexity.

For more information see [github repository](https://github.com/deepmind/acme)."""

# Get the version from metadata.
version = _metadata.__version__

# If we're releasing a nightly/dev version append to the version string.
if '--nightly' in sys.argv:
  sys.argv.remove('--nightly')
  version += '.dev' + datetime.datetime.now().strftime('%Y%m%d')


class BuildPy(setuptools.command.build_py.build_py):

  def run(self):
    generate_requirements_file()
    setuptools.command.build_py.build_py.run(self)


class Develop(setuptools.command.develop.develop):

  def run(self):
    generate_requirements_file()
    setuptools.command.develop.develop.run(self)

cmdclass = {
    'build_py': BuildPy,
    'develop': Develop,
}

setup(
    name='dm-acme',
    version=version,
    cmdclass=cmdclass,
    description='A Python library for Reinforcement Learning.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='DeepMind',
    license='Apache License, Version 2.0',
    keywords='reinforcement-learning python machine learning',
    packages=find_packages(),
    package_data={'': ['requirements.txt']},
    include_package_data=True,
    install_requires=core_requirements,
    extras_require={
        'jax': jax_requirements,
        'tf': tf_requirements,
        'testing': testing_requirements,
        'envs': envs_requirements,
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for base."""

import copy
import pickle

from absl.testing import absltest
from acme.testing import fakes
from acme.wrappers import base


class BaseTest(absltest.TestCase):

  def test_pickle_unpickle(self):
    test_env = base.EnvironmentWrapper(environment=fakes.DiscreteEnvironment())

    test_env_pickled = pickle.dumps(test_env)
    test_env_restored = pickle.loads(test_env_pickled)
    self.assertEqual(
        test_env.observation_spec(),
        test_env_restored.observation_spec(),
    )

  def test_deepcopy(self):
    test_env = base.EnvironmentWrapper(environment=fakes.DiscreteEnvironment())
    copied_env = copy.deepcopy(test_env)
    del copied_env

if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme launchpad utilities."""

from absl.testing import absltest

from acme.utils import lp_utils


class EnvironmentLoopTest(absltest.TestCase):

  def test_partial_kwargs(self):

    def foo(a, b, c=2):
      return a, b, c

    def bar(a, b):
      return a, b

    # Override the default values. The last two should be no-ops.
    foo1 = lp_utils.partial_kwargs(foo, c=1)
    foo2 = lp_utils.partial_kwargs(foo)
    bar1 = lp_utils.partial_kwargs(bar)

    # Check that we raise errors on overriding kwargs with no default values
    with self.assertRaises(ValueError):
      lp_utils.partial_kwargs(foo, a=2)

    # CHeck the we raise if we try to override a kwarg that doesn't exist.
    with self.assertRaises(ValueError):
      lp_utils.partial_kwargs(foo, d=2)

    # Make sure we get back the correct values.
    self.assertEqual(foo1(1, 2), (1, 2, 1))
    self.assertEqual(foo2(1, 2), (1, 2, 2))
    self.assertEqual(bar1(1, 2), (1, 2))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adders that use Reverb (github.com/deepmind/reverb) as a backend."""

import abc
import time
from typing import Callable, Iterable, Mapping, NamedTuple, Optional, Sized, Union, Tuple

from absl import logging
from acme import specs
from acme import types
from acme.adders import base
import dm_env
import numpy as np
import reverb
import tensorflow as tf
import tree

DEFAULT_PRIORITY_TABLE = 'priority_table'
_MIN_WRITER_LIFESPAN_SECONDS = 60
StartOfEpisodeType = Union[bool, specs.Array, tf.Tensor, tf.TensorSpec,
                           Tuple[()]]


# TODO(b/188510142): Delete Step.
class Step(NamedTuple):
  """Step class used internally for reverb adders."""
  observation: types.NestedArray
  action: types.NestedArray
  reward: types.NestedArray
  discount: types.NestedArray
  start_of_episode: StartOfEpisodeType
  extras: types.NestedArray = ()


# TODO(b/188510142): Replace with proper Trajectory class.
Trajectory = Step


class PriorityFnInput(NamedTuple):
  """The input to a priority function consisting of stacked steps."""
  observations: types.NestedArray
  actions: types.NestedArray
  rewards: types.NestedArray
  discounts: types.NestedArray
  start_of_episode: types.NestedArray
  extras: types.NestedArray


# Define the type of a priority function and the mapping from table to function.
PriorityFn = Callable[['PriorityFnInput'], float]
PriorityFnMapping = Mapping[str, Optional[PriorityFn]]


def spec_like_to_tensor_spec(paths: Iterable[str], spec: specs.Array):
  return tf.TensorSpec.from_spec(spec, name='/'.join(str(p) for p in paths))


class ReverbAdder(base.Adder):
  """Base class for Reverb adders."""

  def __init__(
      self,
      client: reverb.Client,
      max_sequence_length: int,
      max_in_flight_items: int,
      delta_encoded: bool = False,
      priority_fns: Optional[PriorityFnMapping] = None,
      get_signature_timeout_ms: int = 300_000,
  ):
    """Initialize a ReverbAdder instance.

    Args:
      client: A client to the Reverb backend.
      max_sequence_length: The maximum length of sequences (corresponding to the
        number of observations) that can be added to replay.
      max_in_flight_items: The maximum number of items allowed to be "in flight"
        at the same time. See `block_until_num_items` in
        `reverb.TrajectoryWriter.flush` for more info.
      delta_encoded: If `True` (False by default) enables delta encoding, see
        `Client` for more information.
      priority_fns: A mapping from table names to priority functions; if
        omitted, all transitions/steps/sequences are given uniform priorities
        (1.0) and placed in DEFAULT_PRIORITY_TABLE.
      get_signature_timeout_ms: time before timeout in fetching the signature
        from the reverb server.
    """
    if priority_fns:
      priority_fns = dict(priority_fns)
    else:
      priority_fns = {DEFAULT_PRIORITY_TABLE: None}

    self._client = client
    self._priority_fns = priority_fns
    self._max_sequence_length = max_sequence_length
    self._delta_encoded = delta_encoded
    # TODO(b/206629159): Remove this.
    self._max_in_flight_items = max_in_flight_items
    self._add_first_called = False

    # This is exposed as the _writer property in such a way that it will create
    # a new writer automatically whenever the internal __writer is None. Users
    # should ONLY ever interact with self._writer.
    self.__writer = None
    # Every time a new writer is created, it must fetch the signature from the
    # Reverb server. If this is set too low it can crash the adders in a
    # distributed setup where the replay may take a while to spin up.
    self._get_signature_timeout_ms = get_signature_timeout_ms

  def __del__(self):
    if self.__writer is not None:
      timeout_ms = 10_000
      # Try flush all appended data before closing to avoid loss of experience.
      try:
        self.__writer.flush(0, timeout_ms=timeout_ms)
      except reverb.DeadlineExceededError as e:
        logging.error(
            'Timeout (%d ms) exceeded when flushing the writer before '
            'deleting it. Caught Reverb exception: %s', timeout_ms, str(e))
      self.__writer.close()

  @property
  def _writer(self) -> reverb.TrajectoryWriter:
    if self.__writer is None:
      self.__writer = self._client.trajectory_writer(
          num_keep_alive_refs=self._max_sequence_length,
          get_signature_timeout_ms=self._get_signature_timeout_ms)
      self._writer_created_timestamp = time.time()
    return self.__writer

  def add_priority_table(self, table_name: str,
                         priority_fn: Optional[PriorityFn]):
    if table_name in self._priority_fns:
      raise ValueError(
          f'A priority function already exists for {table_name}. '
          f'Existing tables: {", ".join(self._priority_fns.keys())}.'
      )
    self._priority_fns[table_name] = priority_fn

  def reset(self, timeout_ms: Optional[int] = None):
    """Resets the adder's buffer."""
    if self.__writer:
      # Flush all appended data and clear the buffers.
      self.__writer.end_episode(clear_buffers=True, timeout_ms=timeout_ms)

      # Create a new writer unless the current one is too young.
      # This is to reduce the relative overhead of creating a new Reverb writer.
      if (time.time() - self._writer_created_timestamp >
          _MIN_WRITER_LIFESPAN_SECONDS):
        self.__writer = None
    self._add_first_called = False

  def add_first(self, timestep: dm_env.TimeStep):
    """Record the first observation of a trajectory."""
    if not timestep.first():
      raise ValueError('adder.add_first with an initial timestep (i.e. one for '
                       'which timestep.first() is True')

    # Record the next observation but leave the history buffer row open by
    # passing `partial_step=True`.
    self._writer.append(dict(observation=timestep.observation,
                             start_of_episode=timestep.first()),
                        partial_step=True)
    self._add_first_called = True

  def add(self,
          action: types.NestedArray,
          next_timestep: dm_env.TimeStep,
          extras: types.NestedArray = ()):
    """Record an action and the following timestep."""

    if not self._add_first_called:
      raise ValueError('adder.add_first must be called before adder.add.')

    # Add the timestep to the buffer.
    has_extras = (len(extras) > 0 if isinstance(extras, Sized)  # pylint: disable=g-explicit-length-test
                  else extras is not None)
    current_step = dict(
        # Observation was passed at the previous add call.
        action=action,
        reward=next_timestep.reward,
        discount=next_timestep.discount,
        # Start of episode indicator was passed at the previous add call.
        **({'extras': extras} if has_extras else {})
    )
    self._writer.append(current_step)

    # Record the next observation and write.
    self._writer.append(
        dict(
            observation=next_timestep.observation,
            start_of_episode=next_timestep.first()),
        partial_step=True)
    self._write()

    if next_timestep.last():
      # Complete the row by appending zeros to remaining open fields.
      # TODO(b/183945808): remove this when fields are no longer expected to be
      # of equal length on the learner side.
      dummy_step = tree.map_structure(np.zeros_like, current_step)
      self._writer.append(dummy_step)
      self._write_last()
      self.reset()

  @classmethod
  def signature(cls, environment_spec: specs.EnvironmentSpec,
                extras_spec: types.NestedSpec = ()):
    """This is a helper method for generating signatures for Reverb tables.

    Signatures are useful for validating data types and shapes, see Reverb's
    documentation for details on how they are used.

    Args:
      environment_spec: A `specs.EnvironmentSpec` whose fields are nested
        structures with leaf nodes that have `.shape` and `.dtype` attributes.
        This should come from the environment that will be used to generate
        the data inserted into the Reverb table.
      extras_spec: A nested structure with leaf nodes that have `.shape` and
        `.dtype` attributes. The structure (and shapes/dtypes) of this must
        be the same as the `extras` passed into `ReverbAdder.add`.

    Returns:
      A `Step` whose leaf nodes are `tf.TensorSpec` objects.
    """
    spec_step = Step(
        observation=environment_spec.observations,
        action=environment_spec.actions,
        reward=environment_spec.rewards,
        discount=environment_spec.discounts,
        start_of_episode=specs.Array(shape=(), dtype=bool),
        extras=extras_spec)
    return tree.map_structure_with_path(spec_like_to_tensor_spec, spec_step)

  @abc.abstractmethod
  def _write(self):
    """Write data to replay from the buffer."""

  @abc.abstractmethod
  def _write_last(self):
    """Write data to replay from the buffer."""



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local R2D2 agents, using JAX."""

import functools
from typing import Callable, Optional

from acme import specs
from acme.agents.jax.r2d2 import builder
from acme.agents.jax.r2d2 import config as r2d2_config
from acme.agents.jax.r2d2 import networks as r2d2_networks
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env
import haiku as hk
import rlax


NetworkFactory = Callable[[specs.EnvironmentSpec], r2d2_networks.R2D2Networks]


class DistributedR2D2FromConfig(distributed_layout.DistributedLayout):
  """Distributed R2D2 agents from config."""

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      environment_spec: specs.EnvironmentSpec,
      network_factory: NetworkFactory,
      config: r2d2_config.R2D2Config,
      seed: int,
      num_actors: int,
      workdir: Optional[str] = '~/acme',
      device_prefetch: bool = False,
      log_to_bigtable: bool = True,
      log_every: float = 10.0,
  ):
    logger_fn = functools.partial(loggers.make_default_logger,
                                  'learner', log_to_bigtable,
                                  time_delta=log_every, asynchronous=True,
                                  serialize_fn=utils.fetch_devicearray,
                                  steps_key='learner_steps')
    r2d2_builder = builder.R2D2Builder(
        networks=network_factory(environment_spec),
        config=config,
        logger_fn=logger_fn)
    policy_network_factory = (
        lambda n: r2d2_networks.make_behavior_policy(n, config))
    evaluator_policy_network_factory = (
        lambda n: r2d2_networks.make_behavior_policy(n, config, True))
    super().__init__(
        seed=seed,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=r2d2_builder,
        policy_network=policy_network_factory,
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=r2d2_builder,
                policy_factory=evaluator_policy_network_factory,
                log_to_bigtable=log_to_bigtable)
        ],
        num_actors=num_actors,
        environment_spec=environment_spec,
        device_prefetch=device_prefetch,
        log_to_bigtable=log_to_bigtable,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            log_to_bigtable, log_every),
        prefetch_size=config.prefetch_size,
        workdir=workdir)


class DistributedR2D2(DistributedR2D2FromConfig):
  """Distributed R2D2 agent."""

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      environment_spec: specs.EnvironmentSpec,
      forward: hk.Transformed,
      unroll: hk.Transformed,
      initial_state: hk.Transformed,
      num_actors: int,
      num_caches: int = 1,
      burn_in_length: int = 40,
      trace_length: int = 80,
      sequence_period: int = 40,
      batch_size: int = 64,
      prefetch_size: int = 2,
      target_update_period: int = 2500,
      samples_per_insert: float = 0.,
      min_replay_size: int = 1000,
      max_replay_size: int = 100_000,
      importance_sampling_exponent: float = 0.6,
      priority_exponent: float = 0.9,
      max_priority_weight: float = 0.9,
      bootstrap_n: int = 5,
      clip_rewards: bool = False,
      tx_pair: rlax.TxPair = rlax.SIGNED_HYPERBOLIC_PAIR,
      learning_rate: float = 1e-3,
      evaluator_epsilon: float = 0.,
      discount: float = 0.997,
      variable_update_period: int = 400,
      seed: int = 1,
    ):
    config = r2d2_config.R2D2Config(
        discount=discount,
        target_update_period=target_update_period,
        evaluation_epsilon=evaluator_epsilon,
        burn_in_length=burn_in_length,
        trace_length=trace_length,
        sequence_period=sequence_period,
        learning_rate=learning_rate,
        bootstrap_n=bootstrap_n,
        clip_rewards=clip_rewards,
        tx_pair=tx_pair,
        samples_per_insert=samples_per_insert,
        min_replay_size=min_replay_size,
        max_replay_size=max_replay_size,
        batch_size=batch_size,
        prefetch_size=prefetch_size,
        importance_sampling_exponent=importance_sampling_exponent,
        priority_exponent=priority_exponent,
        max_priority_weight=max_priority_weight,
    )
    network_factory = functools.partial(
        r2d2_networks.make_networks,
        forward_fn=forward,
        initial_state_fn=initial_state,
        unroll_fn=unroll,
        batch_size=batch_size)
    super().__init__(
        seed=seed,
        environment_factory=environment_factory,
        environment_spec=environment_spec,
        network_factory=network_factory,
        config=config,
        num_actors=num_actors,
    )


class R2D2(local_layout.LocalLayout):
  """Local agent for r2d2.

  This implements a single-process R2D2 agent. This is a simple Q-learning
  algorithm that generates data via a (epsilon-greedy) behavior policy, inserts
  trajectories into a replay buffer, and periodically updates its policy by
  sampling these transitions using prioritization.
  """

  def __init__(
      self,
      spec: specs.EnvironmentSpec,
      networks: r2d2_networks.R2D2Networks,
      config: r2d2_config.R2D2Config,
      seed: int,
      workdir: Optional[str] = '~/acme',
      counter: Optional[counting.Counter] = None,
  ):
    r2d2_builder = builder.R2D2Builder(networks, config)
    super().__init__(
        seed=seed,
        environment_spec=spec,
        builder=r2d2_builder,
        networks=networks,
        policy_network=r2d2_networks.make_behavior_policy(networks, config),
        workdir=workdir,
        min_replay_size=32 * config.sequence_period,
        samples_per_insert=1.,
        batch_size=config.batch_size,
        num_sgd_steps_per_step=config.sequence_period,
        counter=counter,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the D4PG agent."""

from typing import Sequence

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.jax import d4pg
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.testing import fakes
from acme.utils import counting
import haiku as hk
import jax.numpy as jnp
import numpy as np


def make_networks(
    spec: specs.EnvironmentSpec,
    policy_layer_sizes: Sequence[int] = (300, 200),
    critic_layer_sizes: Sequence[int] = (400, 300),
    vmin: float = -150.,
    vmax: float = 150.,
    num_atoms: int = 51,
) -> d4pg.D4PGNetworks:
  """Creates networks used by the agent."""

  action_spec = spec.actions

  num_dimensions = np.prod(action_spec.shape, dtype=int)
  critic_atoms = jnp.linspace(vmin, vmax, num_atoms)

  def _actor_fn(obs):
    network = hk.Sequential([
        utils.batch_concat,
        networks_lib.LayerNormMLP(policy_layer_sizes, activate_final=True),
        networks_lib.NearZeroInitializedLinear(num_dimensions),
        networks_lib.TanhToSpec(action_spec),
    ])
    return network(obs)

  def _critic_fn(obs, action):
    network = hk.Sequential([
        utils.batch_concat,
        networks_lib.LayerNormMLP(layer_sizes=[*critic_layer_sizes, num_atoms]),
    ])
    value = network([obs, action])
    return value, critic_atoms

  policy = hk.without_apply_rng(hk.transform(_actor_fn))
  critic = hk.without_apply_rng(hk.transform(_critic_fn))

  # Create dummy observations and actions to create network parameters.
  dummy_action = utils.zeros_like(spec.actions)
  dummy_obs = utils.zeros_like(spec.observations)
  dummy_action = utils.add_batch_dim(dummy_action)
  dummy_obs = utils.add_batch_dim(dummy_obs)

  return d4pg.D4PGNetworks(
      policy_network=networks_lib.FeedForwardNetwork(
          lambda rng: policy.init(rng, dummy_obs), policy.apply),
      critic_network=networks_lib.FeedForwardNetwork(
          lambda rng: critic.init(rng, dummy_obs, dummy_action), critic.apply))


class D4PGTest(absltest.TestCase):

  def test_d4pg(self):
    # Create a fake environment to test with.
    environment = fakes.ContinuousEnvironment(
        episode_length=10, action_dim=3, observation_dim=5, bounded=True)
    spec = specs.make_environment_spec(environment)

    # Create the networks.
    networks = make_networks(spec)

    config = d4pg.D4PGConfig(
        batch_size=10,
        samples_per_insert=2,
        min_replay_size=10,
        samples_per_insert_tolerance_rate=float('inf'))
    counter = counting.Counter()
    agent = d4pg.D4PG(spec, networks, config=config, random_seed=0,
                      counter=counter)

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent, counter=counter)
    loop.run(num_episodes=2)


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Acme loggers."""

from acme.utils.loggers.aggregators import Dispatcher
from acme.utils.loggers.asynchronous import AsyncLogger
from acme.utils.loggers.auto_close import AutoCloseLogger
from acme.utils.loggers.base import Logger
from acme.utils.loggers.base import LoggingData
from acme.utils.loggers.base import NoOpLogger
from acme.utils.loggers.base import to_numpy
from acme.utils.loggers.constant import ConstantLogger
from acme.utils.loggers.csv import CSVLogger
from acme.utils.loggers.dataframe import InMemoryLogger
from acme.utils.loggers.filters import GatedFilter
from acme.utils.loggers.filters import KeyFilter
from acme.utils.loggers.filters import NoneFilter
from acme.utils.loggers.filters import TimeFilter
from acme.utils.loggers.default import make_default_logger  # pylint: disable=g-bad-import-order
from acme.utils.loggers.terminal import TerminalLogger

# Internal imports.



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for acme.tf.networks.distributional."""

from absl.testing import absltest
from absl.testing import parameterized
from acme.tf.networks import distributional
import numpy as np
from numpy import testing as npt


class DistributionalTest(parameterized.TestCase):

  @parameterized.parameters(
      ((2, 3), (), (), 5, (2, 5)),
      ((2, 3), (4, 1), (1, 5), 6, (2, 4, 5, 6)),
      )
  def test_discrete_valued_head(
      self,
      input_shape,
      vmin_shape,
      vmax_shape,
      num_atoms,
      expected_logits_shape):

    vmin = np.zeros(vmin_shape, np.float)
    vmax = np.ones(vmax_shape, np.float)
    head = distributional.DiscreteValuedHead(
        vmin=vmin,
        vmax=vmax,
        num_atoms=num_atoms)
    input_array = np.zeros(input_shape, dtype=np.float)
    output_distribution = head(input_array)
    self.assertEqual(output_distribution.logits_parameter().shape,
                     expected_logits_shape)

    values = output_distribution._values

    # Can't do assert_allclose(values[..., 0], vmin), because the args may
    # have broadcast-compatible but unequal shapes. Do the following instead:
    npt.assert_allclose(values[..., 0] - vmin, np.zeros_like(values[..., 0]))
    npt.assert_allclose(values[..., -1] - vmax, np.zeros_like(values[..., -1]))

    # Check that values are monotonically increasing.
    intervals = values[..., 1:] - values[..., :-1]
    npt.assert_array_less(np.zeros_like(intervals), intervals)

    # Check that the values are equally spaced.
    npt.assert_allclose(intervals[..., 1:] - intervals[..., :1],
                        np.zeros_like(intervals[..., 1:]),
                        atol=1e-7)


if __name__ == '__main__':
  absltest.main()
